import typing as t
from enum import Enum
from pathlib import Path

from pydantic import Field, PrivateAttr, PositiveInt

from pipelime.piper import PipelimeCommand, PiperPortType

if t.TYPE_CHECKING:
    from pipelime.piper.graph import DAGNodesGraph


class WatcherBackend(Enum):
    """The watcher backend to use."""

    RICH = "rich"
    TQDM = "tqdm"

    def listener_key(self) -> str:
        return {
            WatcherBackend.RICH: "RICH_TABLE",
            WatcherBackend.TQDM: "TQDM_BARS",
        }[self]


class PiperGraphCommandBase(PipelimeCommand):
    """Base class for piper-aware commands."""

    nodes: t.Mapping[
        str, t.Union[PipelimeCommand, t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]]]
    ] = Field(
        ...,
        alias="n",
        description=(
            "A DAG of commands as a `<node>: <command>` mapping. The command can be a "
            "`<name>: <args>` mapping, where `<name>` is `pipe`, `clone`, `split` etc, "
            "while `<args>` is a mapping of its arguments."
        ),
        piper_port=PiperPortType.INPUT,
    )
    include: t.Union[str, t.Sequence[str], None] = Field(
        None, alias="i", description="Nodes not in this list are not run."
    )
    exclude: t.Union[str, t.Sequence[str], None] = Field(
        None, alias="e", description="Nodes in this list are not run."
    )

    _piper_graph: t.Optional["DAGNodesGraph"] = PrivateAttr(None)

    @property
    def piper_graph(self) -> "DAGNodesGraph":
        from pipelime.piper.model import DAGModel, NodesDefinition
        from pipelime.piper.graph import DAGNodesGraph

        if not self._piper_graph:
            inc_n = [self.include] if isinstance(self.include, str) else self.include
            exc_n = [self.exclude] if isinstance(self.exclude, str) else self.exclude

            def _node_to_run(node: str) -> bool:
                return (inc_n is None or node in inc_n) and (
                    exc_n is None or node not in exc_n
                )

            nodes = {
                name: cmd for name, cmd in self.nodes.items() if _node_to_run(name)
            }
            dag = DAGModel(nodes=NodesDefinition.create(nodes))
            self._piper_graph = DAGNodesGraph.build_nodes_graph(
                dag, **self._nodes_graph_building_kwargs()
            )
        return self._piper_graph

    def _nodes_graph_building_kwargs(self) -> t.Mapping[str, t.Any]:
        """Subclasses should returns the extra kwargs options for
        ``DAGNodesGraph.build_nodes_graph``
        """
        return {}


class GraphPortForwardingCommand(PiperGraphCommandBase):
    """A command that uses the root and leaf data nodes of the internal DAG
    as its own I/O ports."""

    def get_inputs(self) -> t.Dict[str, t.Any]:
        return {
            self.piper_graph.get_input_port_name(x): x.path
            for x in self.piper_graph.input_data_nodes
        }

    def get_outputs(self) -> t.Dict[str, t.Any]:
        return {
            self.piper_graph.get_output_port_name(x): x.path
            for x in self.piper_graph.output_data_nodes
        }


class RunCommand(GraphPortForwardingCommand, title="run"):
    """Executes a DAG of pipelime commands."""

    token: t.Optional[str] = Field(
        None,
        alias="t",
        description=(
            "The execution token. If not specified, a new token will be generated."
        ),
    )
    watch: t.Union[bool, WatcherBackend, None] = Field(
        None,
        alias="w",
        description=(
            "Monitor the execution in the current console. "
            "Defaults to True if no token is provided, False othrewise. "
            "If a string is provided, it is used as the name of the watcher backend."
        ),
    )

    def run(self):
        import uuid

        from pipelime.piper.executors.factory import NodesGraphExecutorFactory

        if self._piper.active:
            # nested graph should disable the default watcher
            # and forward the token of the parent graph
            watch = False
            token = self._piper.token
            message = f"{self._piper.node} DAG"
            prefix = f"{self._piper.node}."
        else:
            watch = not self.token if self.watch is None else self.watch
            token = self.token or uuid.uuid1().hex
            message = "Main DAG"
            prefix = ""

            # activate piper, so that this node will send updates to the watchers
            self.set_piper_info(token=token, node=message)

        executor = NodesGraphExecutorFactory.get_executor(
            watch=watch if isinstance(watch, bool) else watch.listener_key(),
            node_prefix=prefix,
            task=self.create_task(
                total=self.piper_graph.num_operation_nodes, message=message
            ),
        )
        if not executor.exec(self.piper_graph, token=token):
            raise RuntimeError("Piper execution failed")


class DrawCommand(PiperGraphCommandBase, title="draw"):
    """Draws a pipelime DAG."""

    class DrawBackendChoice(Enum):
        GRAPHVIZ = "graphviz"
        MERMAID = "mermaid"

    class EllipsesChoice(Enum):
        START = "start"
        MIDDLE = "middle"
        END = "end"
        REGEX = "regex"

    output: t.Optional[Path] = Field(
        None,
        alias="o",
        description=(
            "The output file. If not specified, the graph will be shown in a window."
        ),
        piper_port=PiperPortType.OUTPUT,
    )
    raw_output: t.Optional[str] = Field(
        None,
        description="Raw graph representation",
        exclude=True,
        repr=False,
        piper_port=PiperPortType.OUTPUT,
    )
    backend: DrawBackendChoice = Field(
        DrawBackendChoice.GRAPHVIZ, alias="b", description="The graph backend to use."
    )
    open: bool = Field(
        False,
        description=(
            "If `output` has been set, open the image file in the default viewer."
        ),
    )

    data_max_width: t.Union[PositiveInt, str, None] = Field(
        None,
        alias="m",
        description=(
            "If an int is given, it is the maximum data node name length. "
            "If a string is given, it is matched against the node name and "
            "replaced with ellipses. Note that the search starts from "
            "the last character if `ellipsis_position` is `start`."
        ),
    )
    ellipsis_position: EllipsesChoice = Field(
        EllipsesChoice.MIDDLE,
        alias="ep",
        description="Where ellipses should be put if the data node name is too long.",
    )
    show_command_names: bool = Field(
        False, alias="c", description="Show command names instead of node names."
    )
    extra_args: t.Optional[t.Mapping[str, t.Any]] = Field(
        None, alias="x", description="Extra arguments to pass to the backend."
    )
    raw: bool = Field(
        False, alias="r", description="Show the raw graph representation."
    )

    def _nodes_graph_building_kwargs(self) -> t.Mapping[str, t.Any]:
        return {
            "data_max_width": self.data_max_width,
            "show_command_name": self.show_command_names,
            "ellipsis_position": self.ellipsis_position.value,
        }

    def run(self):
        import os
        import platform
        import subprocess

        from pipelime.piper.drawing.factory import NodesGraphDrawerFactory

        def start_file(filename: str):
            if platform.system() == "Darwin":  # macOS
                subprocess.call(("open", filename))
            elif platform.system() == "Windows":  # Windows
                os.startfile(filename)
            else:  # linux variants #TODO: verify!
                subprocess.call(("xdg-open", filename))

        graph = self.piper_graph
        drawer = NodesGraphDrawerFactory.create(self.backend.value)

        extra = self.extra_args or {}

        # raw graph representation
        if self.raw:
            self.raw_output = drawer.representation(graph)
        else:
            # Show or Write
            if self.output is not None:
                drawer.export(graph, str(self.output), **extra)
                if self.open:
                    start_file(str(self.output))
            else:
                from PIL import Image

                graph_image = drawer.draw(graph=graph, **extra)
                img = Image.fromarray(graph_image, "RGB")
                img.show("Graph")


class WatchCommand(PipelimeCommand, title="watch"):
    token: str = Field(
        ..., alias="t", description="The token of the DAG you want to monitor."
    )
    watcher: WatcherBackend = Field(
        WatcherBackend.TQDM, alias="w", description="The listener to use."
    )

    def run(self):
        from pipelime.piper.progress.listener.base import Listener
        from pipelime.piper.progress.listener.factory import (
            ListenerCallbackFactory,
            ProgressReceiverFactory,
        )
        from pipelime.utils.context_managers import CatchSignals
        from time import sleep

        receiver = ProgressReceiverFactory.get_receiver(self.token)
        callback = ListenerCallbackFactory.get_callback(self.watcher.listener_key())
        listener = Listener(receiver, callback)
        listener.start()

        with CatchSignals() as catcher:
            while not catcher.interrupted:
                sleep(0.01)

        listener.stop()



import typing as t


class SubCommands:
    HELP = (("help", "h"), "show help for a command, an operator or a stage")
    LIST = (("list", "l", "ll", "ls"), "list all commands, operators or stages")
    LIST_CMDS = (
        (
            "list-commands",
            "list-cmds",
            "list-cmd",
            "lc",
            "lcmd",
            "lcmds",
            "ls-cmds",
            "ls-cmd",
            "lsc",
        ),
        "list all commands",
    )
    LIST_OPS = (
        (
            "list-operators",
            "list-ops",
            "list-op",
            "lo",
            "lop",
            "lops",
            "ls-ops",
            "ls-op",
            "lso",
        ),
        "list all operators",
    )
    LIST_STGS = (
        (
            "list-stages",
            "list-stgs",
            "list-stg",
            "lst",
            "lstg",
            "lstgs",
            "ls-stgs",
            "ls-stg",
            "lss",
        ),
        "list all stages",
    )
    AUDIT = (("audit", "a"), "inspects configuration and context")
    WIZARD = (
        ("wizard", "wiz", "w"),
        "interactive wizard to fill command and general model configurations",
    )
    EXEC = (
        ("exec", "exe", "x", "e"),
        "execute a configuration where the command is the top-level key",
    )

    ALL_SUBC = [HELP, LIST, LIST_CMDS, LIST_OPS, LIST_STGS, AUDIT, WIZARD, EXEC]

    @classmethod
    def get_help(cls) -> str:
        return "\n".join(
            ["\b\nSPECIAL SUBCOMMANDS:"]
            + [
                el
                for sc in cls.ALL_SUBC
                for el in ("- " + ", ".join(sc[0]), f"    {sc[1]}")
            ]
        )

    @classmethod
    def get_autocompletions(cls) -> t.List[t.Tuple[str, str]]:
        return [(sc[0][0], sc[1]) for sc in cls.ALL_SUBC]



from pipelime.sequences.pipes.base import PipedSequenceBase
import pipelime.sequences.pipes.operations
import pipelime.sequences.pipes.mapping
import pipelime.sequences.pipes.writers
import pipelime.sequences.pipes.validation



import typing as t
import numpy as np

import pydantic as pyd

import pipelime.sequences as pls


@pls.source_sequence
class ToyDataset(
    pls.SamplesSequence, title="toy_dataset", arbitrary_types_allowed=True
):
    """A fake sequence of generated samples."""

    length: pyd.PositiveInt = pyd.Field(
        ..., description="The number of samples to generate."
    )
    with_images: bool = pyd.Field(True, description="Whether to generate images.")
    with_masks: bool = pyd.Field(
        True, description="Whether to generate masks with object labels."
    )
    with_instances: bool = pyd.Field(
        True, description="Whether to generate images with object indexes."
    )
    with_objects: bool = pyd.Field(True, description="Whether to generate objects.")
    with_bboxes: bool = pyd.Field(
        True, description="Whether to generate objects' bboxes."
    )
    with_kpts: bool = pyd.Field(
        True, description="Whether to generate objects' keypoints."
    )
    image_size: t.Union[
        pyd.PositiveInt, t.Tuple[pyd.PositiveInt, pyd.PositiveInt]
    ] = pyd.Field(256, description="The size of the generated images.")
    key_format: str = pyd.Field(
        "*",
        description=(
            "The sample key format. Any `*` will be replaced with the "
            "base key name, eg, `my_*_key` on [`image`, `mask`] generates "
            "`my_image_key` and `my_mask_key`. If no `*` is found, the string is "
            "suffixed to the base key name, ie, `MyKey` on `image` gives "
            "`imageMyKey`. If empty, the base key name will be used as-is."
        ),
    )
    max_labels: pyd.NonNegativeInt = pyd.Field(
        5, description="The maximum number assigned to object labels in the dataset."
    )
    objects_range: t.Tuple[pyd.NonNegativeInt, pyd.NonNegativeInt] = pyd.Field(
        (1, 5), description="The (min, max) number of objects in each sample."
    )

    seed: t.Optional[int] = pyd.Field(None, description="The optional random seed.")

    _sample_cache: t.Dict[int, pls.Sample] = pyd.PrivateAttr(default_factory=dict)
    _rnd_gen: np.random.Generator = pyd.PrivateAttr(None)

    @pyd.validator("key_format")
    def validate_key_format(cls, v):
        if "*" in v:
            return v
        return "*" + v

    def __init__(self, length: int, **data):
        super().__init__(length=length, **data)  # type: ignore

    def size(self) -> int:
        return self.length

    def image_shape(self) -> t.Tuple[int, int]:
        if isinstance(self.image_size, int):
            return (self.image_size, self.image_size)
        return self.image_size

    def get_sample(self, idx: int) -> pls.Sample:
        if idx < 0 or idx >= self.size():
            raise IndexError(f"Sample index `{idx}` is out of range.")
        if idx in self._sample_cache:
            return self._sample_cache[idx]

        # create the random generator here, so that processes do not share the seed
        if self._rnd_gen is None:
            self._rnd_gen = np.random.default_rng(self.seed)

        sample = self._generate_sample(idx)
        self._sample_cache[idx] = sample
        return sample

    def _generate_sample(self, idx) -> pls.Sample:
        import uuid
        import pipelime.items as pli
        import json

        label_key = self.key_format.replace("*", "label")
        bboxes_key = self.key_format.replace("*", "bboxes")
        kpts_key = self.key_format.replace("*", "keypoints")

        metadata = {
            label_key: int(self._rnd_gen.integers(self.max_labels + 1)),
            self.key_format.replace("*", "id"): idx
            + (
                self.seed
                if self.seed is not None
                else int(self._rnd_gen.integers(self.length))
            ),
        }

        objects = []
        if self.with_objects:
            no_objs = self._rnd_gen.integers(
                self.objects_range[0], self.objects_range[1]
            )
            objects = [self._generate_2d_object() for _ in range(no_objs)]
            if self.with_bboxes:
                metadata[bboxes_key] = [self._generate_bbox(obj) for obj in objects]
            if self.with_kpts:
                metadata[kpts_key] = [self._generate_kpts(obj) for obj in objects]

        items: t.Dict[str, pli.Item] = {
            k: pli.PngImageItem(v) for k, v in self._generate_images(objects).items()
        }
        items[self.key_format.replace("*", "metadata")] = pli.YamlMetadataItem(
            json.loads(json.dumps(metadata))
        )
        items[label_key] = pli.TxtNumpyItem(metadata[label_key])
        if self.with_objects and self.with_bboxes:
            items[bboxes_key] = pli.NpyNumpyItem(metadata[bboxes_key])
        if self.with_objects and self.with_kpts:
            items[kpts_key] = pli.PickleItem(metadata[kpts_key])

        return pls.Sample(items)

    def _generate_2d_object(self):
        size = np.array(self.image_shape())
        center = self._rnd_gen.uniform(0.25 * size[0], 0.75 * size[0], (2,))
        random_size = self._rnd_gen.uniform(size * 0.05, size * 0.24, (2,))
        top_left = np.array(
            [center[0] - random_size[0] * 0.5, center[1] - random_size[1] * 0.5]
        )
        bottom_right = np.array(
            [center[0] + random_size[0] * 0.5, center[1] + random_size[1] * 0.5]
        )
        diag = bottom_right - top_left  # type: ignore

        kp0 = np.array([center[0], center[1] - random_size[1] * 0.5])
        kp1 = np.array([center[0], center[1] + random_size[1] * 0.5])

        width = random_size[0]
        height = random_size[1]
        label = int(self._rnd_gen.integers(0, self.max_labels + 1))

        return {
            "center": center,
            "size": random_size,
            "tl": top_left,
            "br": bottom_right,
            "diag": diag,
            "kp0": kp0,
            "kp1": kp1,
            "w": width,
            "h": height,
            "label": label,
        }

    def _generate_bbox(self, obj):
        center = obj["center"]
        return (
            center[0] - obj["w"] * 0.5,
            center[1] - obj["h"] * 0.5,
            center[0] + obj["w"] * 0.5,
            center[1] + obj["h"] * 0.5,
            obj["label"],
        )

    def _generate_kpts(self, obj):
        tl = obj["tl"]
        br = obj["br"]
        diag = obj["diag"]
        label = obj["label"]

        orientation = np.arctan2(diag[1], diag[0])
        orientation2 = np.arctan2(-diag[1], -diag[0])

        scale = 0.5 * np.linalg.norm(diag)

        kp1 = (tl[0], tl[1], -orientation, scale, label, 1)
        kp2 = (br[0], br[1], -orientation2, scale, label, 2)

        return [kp1, kp2]

    def _generate_images(self, objs):
        from PIL import Image, ImageDraw

        image = (
            Image.fromarray(
                self._rnd_gen.integers(
                    0, 256, self.image_shape() + (3,), dtype=np.uint8
                )
            )
            if self.with_images
            else None
        )
        mask = Image.new("L", self.image_shape()) if self.with_masks else None
        instances = Image.new("L", self.image_shape()) if self.with_instances else None

        for index, obj in enumerate(objs):
            label = obj["label"]
            coords = tuple(obj["tl"]), tuple(obj["br"])
            color = tuple(self._rnd_gen.integers(0, 255, (3,)))

            if image:
                ImageDraw.Draw(image).rectangle(coords, fill=color)
            if mask:
                ImageDraw.Draw(mask).rectangle(coords, fill=label + 1)
            if instances:
                ImageDraw.Draw(instances).rectangle(coords, fill=index + 1)

        image_map = {}
        if image:
            image_map[self.key_format.replace("*", "image")] = np.array(image)
        if mask:
            image_map[self.key_format.replace("*", "mask")] = np.array(mask)
        if instances:
            image_map[self.key_format.replace("*", "instances")] = np.array(instances)
        return image_map



from pipelime.items.base import (
    Item,
    UnknownItem,
    SerializationMode,
    set_item_serialization_mode,
    item_serialization_mode,
    set_item_disabled_serialization_modes,
    item_disabled_serialization_modes,
    disable_item_data_cache,
    enable_item_data_cache,
    no_data_cache,
    data_cache,
)

# import and register all items
from pipelime.items.binary_item import BinaryItem
from pipelime.items.pickle_item import PickleItem
from pipelime.items.numpy_item import NumpyItem, NpyNumpyItem, TxtNumpyItem
from pipelime.items.image_item import (
    ImageItem,
    BmpImageItem,
    PngImageItem,
    JpegImageItem,
    TiffImageItem,
)
from pipelime.items.metadata_item import (
    MetadataItem,
    JsonMetadataItem,
    YamlMetadataItem,
    TomlMetadataItem,
)
from pipelime.items.model3d_item import (
    Model3DItem,
    STLModel3DItem,
    OBJModel3DItem,
    PLYModel3DItem,
    OFFModel3DItem,
    GLBModel3DItem,
)



import typing as t
from abc import ABC, abstractmethod


def parse_pipelime_cli(
    command_args: t.Sequence[str],
) -> t.Tuple[t.Mapping[str, t.Any], t.Mapping[str, t.Any]]:
    """Parses a pipelime command line and returns
    a tuple with (config_options, context_options)
    """
    try:
        cli_state = CLIParserHoldState()
        for idx, token in enumerate(command_args):
            cli_state = cli_state.process_token(token)
        cli_state.close()
    except Exception as e:
        try:
            hints = [
                f"Error raised while parsing token {token}",  # type: ignore
                " ".join(command_args),
                "─" * (sum((len(a) + 1) for a in command_args[:idx]))  # type: ignore
                + "⌃" * len(token),  # type: ignore
            ]
        except NameError:  # pragma: no cover
            hints = []
        if isinstance(e, CLIParsingError):
            raise CLIParsingError(*e.message, hints=hints + list(e.hints))
        raise CLIParsingError(str(e), hints=hints) from e
    return cli_state.config_options, cli_state.context_options


class CLIParsingError(Exception):
    """Exception raised when an error occurs during CLI parsing"""

    def __init__(
        self, *args, hints: t.Union[str, t.Sequence[str], None] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.message = args
        self.hints = (
            [hints] if isinstance(hints, str) else ([] if hints is None else hints)
        )

    def rich_print(self):  # pragma: no cover
        from pipelime.cli.pretty_print import print_error, print_warning

        for m in self.args:
            print_error(m)
        for m in self.hints:
            print_warning(m)


class CLISpecialChars:
    @staticmethod
    def assignment():
        return ("=",)

    @staticmethod
    def config():
        # MUST BE SORTED FROM THE LONGEST TO THE SHORTEST
        return ("++", "+")

    @staticmethod
    def context():
        # MUST BE SORTED FROM THE LONGEST TO THE SHORTEST
        return ("@@", "@")

    @staticmethod
    def ctx_start():
        return ("//",)


class CLIParserState(ABC):
    def __init__(
        self,
        cfg_opts: t.Optional[t.Dict[str, t.Any]] = None,
        ctx_opts: t.Optional[t.Dict[str, t.Any]] = None,
        ctx_started: bool = False,
    ):
        self.cfg_opts = cfg_opts or {}
        self.ctx_opts = ctx_opts or {}
        self.ctx_started = ctx_started

    @property
    def config_options(self):
        return self.cfg_opts

    @property
    def context_options(self):
        return self.ctx_opts

    @abstractmethod
    def process_token(self, token: str) -> "CLIParserState":
        pass

    @abstractmethod
    def close(self):
        pass


class CLIParserHoldState(CLIParserState):
    def process_token(self, token: str) -> CLIParserState:
        if token in CLISpecialChars.ctx_start():
            return CLIParserHoldState(self.cfg_opts, self.ctx_opts, True)
        if not self.ctx_started and token.startswith(CLISpecialChars.config()):
            opt, val = self._process_key_arg(token)
            cli_state = CLIParserExpectingCfgValue(
                opt, self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            if val is not None:
                return cli_state.process_token(val)
            return cli_state
        if (
            self.ctx_started and token.startswith(CLISpecialChars.config())
        ) or token.startswith(CLISpecialChars.context()):
            opt, val = self._process_key_arg(token)
            cli_state = CLIParserExpectingCtxValue(
                opt, self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            if val is not None:
                return cli_state.process_token(val)
            return cli_state

        raise CLIParsingError(f"Unexpected token: `{token}`")

    def close(self):
        return

    def _process_key_arg(self, token: str):
        opt, val = token, None
        for char in (  # pragma: no branch
            CLISpecialChars.config() + CLISpecialChars.context()
        ):
            if token.startswith(char):
                opt = token[len(char) :]  # noqa: E203
                break

        for char in CLISpecialChars.assignment():
            if char in opt:
                opt, _, val = opt.partition(char)
                break

        if opt.endswith(".") or ".." in opt or ".[" in opt:
            raise CLIParsingError(
                f"Invalid key path: `{opt}`",
                hints=[
                    "Remember: Bash and other shells want the choixe's "
                    "dollar sign (`$`) escaped! Try with single quotes or backslash.",
                    "For example:",
                    "bash/zsh: +operations.map.$model => '+operations.map.$model'",
                    r"zsh: +operations.map.$model => +operations.map.\$model",
                ],
            )

        return opt, val


class CLIParserExpectingValue(CLIParserState):
    def __init__(
        self,
        key_name: str,
        cfg_opts: t.Optional[t.Dict[str, t.Any]] = None,
        ctx_opts: t.Optional[t.Dict[str, t.Any]] = None,
        ctx_started: bool = False,
    ):
        super().__init__(cfg_opts, ctx_opts, ctx_started)
        self.key_name = key_name

    @abstractmethod
    def target_cfg(self) -> t.Dict[str, t.Any]:
        pass

    def process_token(self, token: str) -> CLIParserState:
        from pipelime.choixe.utils.common import deep_set_

        if token in CLISpecialChars.ctx_start() or token.startswith(
            CLISpecialChars.config() + CLISpecialChars.context()
        ):
            self._set_boolean_flag()
            cli_state = CLIParserHoldState(
                self.cfg_opts, self.ctx_opts, self.ctx_started
            )
            return cli_state.process_token(token)

        deep_set_(
            self.target_cfg(),
            key_path=self.key_name,
            value=self._convert_val(token),
            append=True,
        )
        return CLIParserHoldState(self.cfg_opts, self.ctx_opts, self.ctx_started)

    def close(self):
        self._set_boolean_flag()

    def _set_boolean_flag(self):
        from pipelime.choixe.utils.common import deep_set_

        deep_set_(
            self.target_cfg(),
            key_path=self.key_name,
            value=True,
            append=True,
        )

    def _convert_val(self, val: str):
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False
        if val.lower() in ("none", "null", "nul"):
            return None
        try:
            num = int(val)
            return num
        except ValueError:
            pass
        try:
            num = float(val)
            return num
        except ValueError:
            pass
        if (val.startswith("'") and val.endswith("'")) or (
            val.startswith('"') and val.endswith('"')
        ):
            return val[1:-1]
        return val


class CLIParserExpectingCfgValue(CLIParserExpectingValue):
    def target_cfg(self) -> t.Dict[str, t.Any]:
        return self.cfg_opts


class CLIParserExpectingCtxValue(CLIParserExpectingValue):
    def target_cfg(self) -> t.Dict[str, t.Any]:
        return self.ctx_opts



from pipelime.cli.utils import (
    print_command_op_stage_info,
    print_commands_ops_stages_list,
    pl_print,
)

import typing as t


class PipelimeApp:
    def __init__(
        self,
        *extra_modules: str,
        app_name: t.Optional[str] = None,
        entry_point: t.Optional[str] = None,
        app_description: t.Optional[str] = None,
        app_version: t.Optional[str] = None,
        **extra_args: t.Union[str, t.Sequence[str]],
    ):
        """Create a PipelimeApp instance.

        Args:
            *extra_modules (str): the extra modules to load.
                Defaults to the caller module
            app_name (t.Optional[str], optional): the app name.
                Defaults to the package name.
            entry_point (t.Optional[str], optional): the name of the entry point.
                Defaults to package name.
            app_description (t.Optional[str], optional): a description of the app.
                Defaults to the docstring of the caller.
            app_version (t.Optional[str], optional): the app version.
                Defaults to the Pipelime version.
            **extra_args (str): additional command line arguments. When an argument
                must be specified multiple times, use a list as value.
        """
        from pipelime.utils.inspection import MyCaller

        caller = MyCaller()
        caller_module = caller.module

        if app_name is None:
            app_name = caller.package.capitalize()

        if app_description is None:
            app_description = caller.docstrings

        self._app_name = app_name
        self._entry_point = entry_point
        self._app_description = app_description
        self._app_version = app_version

        extra_modules = extra_modules if extra_modules else (caller_module,)
        self._extra_args = [p for m in extra_modules for p in ("-m", m)]

        for k, v in extra_args.items():
            if isinstance(v, str):
                self._extra_args += [k, v]
            else:
                self._extra_args.extend((p for vv in v for p in (k, vv)))

    def __call__(self):
        from pipelime.cli.main import run_typer_app

        run_typer_app(
            app_name=self._app_name,
            entry_point=self._entry_point,
            app_description=self._app_description,
            version=self._app_version,
            extra_args=self._extra_args,
        )


def run():
    """Runs a default PipelimeApp instance from your module.

    Example:
        ```
        @command
        def my_command():
            ...

        if __name__ == "__main__":
            run()
        ```
    """
    from pipelime.utils.inspection import MyCaller
    from pathlib import Path

    caller = MyCaller()
    caller_path = Path(caller.filename)

    app = PipelimeApp(
        caller.module,
        app_name=caller_path.stem,
        entry_point=f"python {caller_path.name}",
        app_description=caller.docstrings,
        app_version="<not set>",
    )
    app()



from typing import Optional, Union

from pipelime.piper.executors.base import NodesGraphExecutor, WatcherNodesGraphExecutor
from pipelime.piper.executors.naive import NaiveNodesGraphExecutor
from pipelime.piper.progress.tracker.base import TrackedTask


class NodesGraphExecutorFactory:
    CLASS_MAP = {
        "NAIVE": NaiveNodesGraphExecutor,
    }

    DEFAULT_EXECUTOR_TYPE = "NAIVE"

    @classmethod
    def get_executor(
        cls,
        type_: Optional[str] = None,
        node_prefix: str = "",
        watch: Union[bool, str] = False,
        task: Optional[TrackedTask] = None,
        **kwargs,
    ) -> NodesGraphExecutor:
        executor = cls.CLASS_MAP[(type_ or cls.DEFAULT_EXECUTOR_TYPE).upper()](
            node_prefix=node_prefix, **kwargs
        )
        executor.task = task

        if watch:
            executor = WatcherNodesGraphExecutor(
                executor, watch if isinstance(watch, str) else None
            )

        return executor



import json
from pathlib import Path
from typing import Any

import yaml


def load(path: Path) -> Any:
    """Loads an object from a file with a supported markup format.
    Supported formats include:
    - yaml
    - json

    Args:
        path (Path): Path to the file to load.

    Returns:
        Any: The loaded object.
    """
    with open(path, "r") as fd:
        ext = path.suffix
        if ext == ".json":
            return json.load(fd)
        else:
            return yaml.safe_load(fd)


def dump(obj: Any, path: Path) -> None:
    """Dumps an object to a file with a supported markup format.
    Supported formats include:
    - yaml
    - json
    Args:
        obj (Any): The object to dump.
        path (Path): Path to the file to write.
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w") as fd:
        ext = path.suffix
        if ext == ".json":
            json.dump(obj, fd)
        else:
            yaml.safe_dump(obj, fd, sort_keys=False)



import typing as t
from types import ModuleType
from pydantic import BaseModel, ValidationError

if t.TYPE_CHECKING:
    from pipelime.piper.model import PipelimeCommand
    from pipelime.stages import SampleStage


class ActionInfo(BaseModel):
    action: t.Callable
    name: str
    description: str
    classpath: str


class PipelimeSymbolsHelper:
    std_modules = ["pipelime.commands", "pipelime.stages"]
    extra_modules: t.List[str] = []

    cached_modules: t.Dict[str, ModuleType] = {}
    cached_cmds: t.Dict[t.Tuple[str, str], t.Dict] = {}
    cached_seq_ops: t.Dict[t.Tuple[str, str], t.Dict] = {}
    cached_stages: t.Dict[t.Tuple[str, str], t.Dict] = {}

    registered_actions: t.Dict[str, ActionInfo] = {}

    @classmethod
    def complete_name(
        cls,
        is_cmd: bool,
        is_seq_op: bool,
        is_stage: bool,
        *,
        additional_names: t.Sequence[t.Tuple[str, str]] = [],
    ):
        import inspect

        valid_completion_items = (
            list((cls.get_pipelime_commands() if is_cmd else {}).values())
            + list((cls.get_sequence_operators() if is_seq_op else {}).values())
            + list((cls.get_sample_stages() if is_stage else {}).values())
        )
        valid_completion_items = [
            (k, inspect.getdoc(v) or "")
            for elem in valid_completion_items
            for k, v in elem.items()
        ] + list(additional_names)

        def _complete(incomplete: str):
            for name, help_text in valid_completion_items:
                if name.startswith(incomplete):
                    yield (name, help_text)

        return _complete

    @classmethod
    def set_extra_modules(cls, modules: t.Sequence[str]):
        cls.extra_modules = list(modules)

    @classmethod
    def is_cache_valid(cls) -> bool:
        return all(
            m in cls.cached_modules for m in (cls.std_modules + cls.extra_modules)
        )

    @classmethod
    def register_action(cls, name: str, info: ActionInfo):
        if name in cls.registered_actions:
            raise ValueError(f"Action `{name}` already registered")
        cls.registered_actions[name] = info

    @classmethod
    def _symbol_name(cls, symbol):
        from pydantic import BaseModel

        return (
            symbol.__config__.title
            if issubclass(symbol, BaseModel) and symbol.__config__.title
            else symbol.__name__
        )

    @classmethod
    def _warn_double_def(cls, type_, name, first, second):
        from pipelime.cli.pretty_print import (
            print_error,
            print_warning,
            get_model_classpath,
        )

        print_error(f"Found duplicate {type_} `{name}`")
        print_warning(f"Defined as `{get_model_classpath(first)}`")
        print_warning(f"       and `{get_model_classpath(second)}`")
        raise ValueError(f"Duplicate {type_} `{name}`")

    @classmethod
    def _load_symbols(cls, base_cls: t.Type, symbol_type: str):
        import inspect

        all_syms = {}
        for module_name, module_ in cls.cached_modules.items():
            module_symbols = tuple(
                (cls._symbol_name(sym_cls), sym_cls)
                for _, sym_cls in inspect.getmembers(
                    module_,
                    lambda v: inspect.isclass(v)
                    and issubclass(v, base_cls)
                    and not inspect.isabstract(v),
                )
            )

            # set module path when loading from file
            if module_name.endswith(".py"):
                for _, sym_cls in module_symbols:
                    sym_cls._classpath = f"{module_name}:{sym_cls.__name__}"

            # check for double symbols in the same module
            for idx, (sym_name, sym_cls) in enumerate(module_symbols):
                for other_sym_name, other_sym_cls in module_symbols[idx + 1 :]:  # noqa
                    if sym_name == other_sym_name:
                        cls._warn_double_def(
                            symbol_type, sym_name, sym_cls, other_sym_cls
                        )

            # check for double commands across modules
            module_symbols = dict(module_symbols)
            for sym_name, sym_cls in module_symbols.items():
                if sym_name in all_syms:
                    cls._warn_double_def(
                        symbol_type, sym_name, sym_cls, all_syms[sym_name]
                    )

            all_syms = {**all_syms, **module_symbols}
        return all_syms

    @classmethod
    def import_everything(cls):
        import pipelime.choixe.utils.imports as pl_imports
        import pipelime.sequences as pl_seq
        from pipelime.piper import PipelimeCommand
        from pipelime.stages import SampleStage

        if not cls.is_cache_valid():
            for module_name in cls.std_modules + cls.extra_modules:
                if module_name not in cls.cached_modules:
                    cls.cached_modules[module_name] = (
                        pl_imports.import_module(module_name)
                    )

            cls.cached_seq_ops = {
                (
                    "Sequence Generator",
                    "Sequence Generators",
                ): pl_seq.SamplesSequence._sources,
                (
                    "Sequence Piped Operation",
                    "Sequence Piped Operations",
                ): pl_seq.SamplesSequence._pipes,
            }

            cls.cached_cmds = {
                ("Pipelime Command", "Pipelime Commands"): cls._load_symbols(
                    PipelimeCommand, "command"
                )
            }

            cls.cached_stages = {
                ("Sample Stage", "Sample Stages"): cls._load_symbols(
                    SampleStage, "stage"
                )
            }

    @classmethod
    def get_pipelime_commands(cls):
        cls.import_everything()
        return cls.cached_cmds

    @classmethod
    def get_sequence_operators(cls):
        cls.import_everything()
        return cls.cached_seq_ops

    @classmethod
    def get_sample_stages(cls):
        cls.import_everything()
        return cls.cached_stages

    @classmethod
    def get_actions(cls):
        return cls.registered_actions

    @classmethod
    def get_symbol(
        cls,
        symbol_path: str,
        base_cls: t.Type,
        symbol_cache: t.Mapping[t.Tuple[str, str], t.Mapping],
    ):
        import pipelime.choixe.utils.imports as pl_imports

        if "." in symbol_path or ":" in symbol_path:
            try:
                imported_symbol = pl_imports.import_symbol(symbol_path)
                if not issubclass(imported_symbol, base_cls):
                    raise ValueError(
                        f"{symbol_path} must derived from {base_cls.__name__}."
                    )

                if ".py:" in symbol_path:
                    imported_symbol._classpath = symbol_path

                return imported_symbol
            except (ImportError, TypeError):
                return None

        for sym_type, sym_dict in symbol_cache.items():
            if symbol_path in sym_dict:
                return (sym_type, sym_dict[symbol_path])
        return None

    @classmethod
    def get_command(cls, command_name: str):
        from pipelime.piper import PipelimeCommand

        sym_cls = cls.get_symbol(
            command_name, PipelimeCommand, cls.get_pipelime_commands()
        )
        if sym_cls is None:
            return None
        if not isinstance(sym_cls, tuple):
            return (("Imported Command", "Imported Commands"), sym_cls)
        return sym_cls

    @classmethod
    def get_operator(cls, operator_name: str):
        for op_type, op_dict in PipelimeSymbolsHelper.get_sequence_operators().items():
            if operator_name in op_dict:
                return (op_type, op_dict[operator_name])
        return None

    @classmethod
    def get_stage(cls, stage_name: str):
        from pipelime.stages import SampleStage

        sym_cls = cls.get_symbol(stage_name, SampleStage, cls.get_sample_stages())
        if sym_cls is None:
            return None
        if not isinstance(sym_cls, tuple):
            return (("Imported Stage", "Imported Stages"), sym_cls)
        return sym_cls

    @classmethod
    def get_action(cls, action_name: str):
        return PipelimeSymbolsHelper.get_actions().get(action_name, None)

    @classmethod
    def show_error_and_help(
        cls, name: str, should_be_cmd: bool, should_be_op: bool, should_be_stage: bool
    ):
        from pipelime.cli.pretty_print import print_error, print_warning, print_info
        from difflib import get_close_matches

        names_list = []
        if should_be_cmd:
            names_list += [
                v2 for v1 in cls.get_pipelime_commands().values() for v2 in v1.keys()
            ]
        if should_be_op:
            names_list += [
                v2 for v1 in cls.get_sequence_operators().values() for v2 in v1.keys()
            ]
        if should_be_stage:
            names_list += [
                v2 for v1 in cls.get_sample_stages().values() for v2 in v1.keys()
            ]
        similar_names = get_close_matches(name, names_list, cutoff=0.3)

        print_error(f"{name} is not a pipelime object and cannot be imported.")
        print_warning("Have you added the module with `--module/-m`?")
        if similar_names:
            print_info(f"Similar entries: {', '.join(similar_names)}")


def _print_info(
    info_cls,
    show_class_path=True,
    show_piper_port=True,
    show_description=True,
    recursive=True,
):
    from pipelime.cli.pretty_print import print_info, print_model_info, _short_line

    if info_cls is not None:
        print_info(f"\n{_short_line()} {info_cls[0][0]}")
        print_model_info(
            info_cls[1],
            show_class_path=show_class_path,
            show_piper_port=show_piper_port,
            show_description=show_description,
            recursive=recursive,
        )


def _unwrap_generic_type(symbol, verbose: bool):
    from pipelime.cli.pretty_print import print_debug
    from rich.markup import escape

    if verbose:
        symstr = escape(repr(symbol))
        for name in (
            "Callable",
            "Tuple",
            "List",
            "Dict",
            "Mapping",
            "Sequence",
            "Union",
            "Literal",
            "Final",
            "ClassVar",
        ):
            symstr = symstr.replace(name, f"[dark_red]{name}[/dark_red]")
        print_debug(f"Unwrapping [not italic]{symstr}[/]")
    all_types = set()
    if t.get_origin(symbol) is None:
        all_types.add(symbol)
    else:
        for arg in t.get_args(symbol):
            all_types.update(_unwrap_generic_type(arg, verbose))
    return all_types


def print_command_op_stage_info(
    command_operator_stage: str, show_description: bool = True, recursive: bool = True
):
    """Prints detailed info about a pipelime command, a sequence operator or a sample
    stage.

    Args:
      command_operator_stage: str:
      show_description: bool:  (Default value = True)
      recursive: bool:  (Default value = True)
    """
    from pipelime.cli.pretty_print import print_info
    from pipelime.choixe.utils.imports import import_symbol

    available_defs = []

    try:
        available_defs.append(
            (
                PipelimeSymbolsHelper.get_stage(command_operator_stage),
                {"show_class_path": True, "show_piper_port": False},
            )
        )
    except ValueError:
        pass

    available_defs.append(
        (
            PipelimeSymbolsHelper.get_operator(command_operator_stage),
            {"show_class_path": False, "show_piper_port": False},
        )
    )

    try:
        available_defs.append(
            (
                PipelimeSymbolsHelper.get_command(command_operator_stage),
                {"show_class_path": True, "show_piper_port": True},
            )
        )
    except ValueError:
        pass

    # Remove None entries, if any
    available_defs = [x for x in available_defs if x[0] is not None]

    if not available_defs:
        # Try a generic import
        try:
            sym = import_symbol(command_operator_stage)
        except ImportError:
            pass
        else:
            sym = _unwrap_generic_type(sym, verbose=True)
            available_defs.extend(
                (
                    (
                        ("Imported Symbol", "Imported Symbols"),
                        s,
                    ),
                    {"show_class_path": True, "show_piper_port": False},
                )
                for s in sym
            )

    if not available_defs:
        PipelimeSymbolsHelper.show_error_and_help(
            command_operator_stage,
            should_be_cmd=True,
            should_be_op=True,
            should_be_stage=True,
        )
        raise ValueError(
            f"{command_operator_stage} is not a pipelime object and cannot be imported."
        )

    if len(available_defs) > 1:
        from rich.prompt import Prompt
        from rich.markup import escape

        print_info("Multiple definitions found!")
        idx = Prompt.ask(
            ">>> Show:\n"
            + "\n".join(
                escape(f"[{i:>2}] {v[0][0][0]} {v[0][1]}")
                for i, v in enumerate(available_defs)
            )
            + "\n[-1] ALL\n>>>",
            choices=["-1"] + list(str(v) for v in range(len(available_defs))),
            default="-1",
        )
        idx = int(idx)
        if idx >= 0:
            available_defs = [available_defs[idx]]

    for d in available_defs:
        _print_info(
            d[0], **d[1], show_description=show_description, recursive=recursive
        )


def _print_details(info, show_class_path, show_piper_port, show_description, recursive):
    from pipelime.cli.pretty_print import _short_line, print_info, print_model_info

    for info_type, info_map in info.items():
        for info_cls in info_map.values():
            print_info(f"\n{_short_line()} {info_type[0]}")
            print_model_info(
                info_cls,
                show_class_path=show_class_path,
                show_piper_port=show_piper_port,
                show_description=show_description,
                recursive=recursive,
            )


def _print_short_help(info, show_class_path, *args, **kwargs):
    from pipelime.cli.pretty_print import (
        _short_line,
        print_info,
        print_models_short_help,
    )

    for info_type, info_map in info.items():
        print_info(f"\n{_short_line()} {info_type[1]}")
        print_models_short_help(
            *[info_cls for info_cls in info_map.values()],
            show_class_path=show_class_path,
        )


def print_commands_ops_stages_list(
    show_details: bool = False,
    *,
    show_cmds: bool = True,
    show_ops: bool = True,
    show_stages: bool = True,
    show_description: bool = True,
    recursive: bool = True,
):
    """Print a list of all available sequence operators and pipelime commands."""

    from pipelime.cli.pretty_print import (
        get_model_classpath,
        print_info,
        print_actions_short_help,
        _short_line,
    )

    def _filter_symbols(smbls):
        if not PipelimeSymbolsHelper.extra_modules:
            return smbls
        filtered_smbls = {}
        for k, v in smbls.items():
            vfilt = {}
            for sym_name, sym_cls in v.items():
                if not get_model_classpath(sym_cls).startswith("pipelime."):
                    vfilt[sym_name] = sym_cls
            if vfilt:
                filtered_smbls[k] = vfilt
        return filtered_smbls

    print_fn = _print_details if show_details else _print_short_help
    if show_cmds:
        print_fn(
            _filter_symbols(PipelimeSymbolsHelper.get_pipelime_commands()),
            show_class_path=True,
            show_piper_port=True,
            show_description=show_description,
            recursive=recursive,
        )
    if show_ops:
        print_fn(
            _filter_symbols(PipelimeSymbolsHelper.get_sequence_operators()),
            show_class_path=False,
            show_piper_port=False,
            show_description=show_description,
            recursive=recursive,
        )
    if show_stages:
        print_fn(
            _filter_symbols(PipelimeSymbolsHelper.get_sample_stages()),
            show_class_path=True,
            show_piper_port=False,
            show_description=show_description,
            recursive=recursive,
        )
        if PipelimeSymbolsHelper.get_actions():
            print_info(f"\n{_short_line()} Actions")
            print_actions_short_help(
                *PipelimeSymbolsHelper.get_actions().values(), show_class_path=True
            )


def print_commands_list(show_details: bool = False):
    """Print a list of all available pipelime commands.

    Args:
      show_details: bool:  (Default value = False)
    """
    print_commands_ops_stages_list(
        show_details=show_details, show_cmds=True, show_ops=False, show_stages=False
    )


def print_sequence_operators_list(show_details: bool = False):
    """Print a list of all available sequence operators.

    Args:
      show_details: bool:  (Default value = False)
    """
    print_commands_ops_stages_list(
        show_details=show_details, show_cmds=False, show_ops=True, show_stages=False
    )


def print_sample_stages_list(show_details: bool = False):
    """Print a list of all available sample stages.

    Args:
      show_details: bool:  (Default value = False)
    """
    print_commands_ops_stages_list(
        show_details=show_details, show_cmds=False, show_ops=False, show_stages=True
    )


def pl_print(
    *objects: t.Any,
    sep: str = " ",
    end: str = "\n",
    file: t.Optional[t.IO[str]] = None,
    flush: bool = False,
):
    """Prints Items, Samples, Sequences, commands, stages, operations...

    Args:
      *objects: t.Any:
      sep: str:  (Default value = " ")
      end: str:  (Default value = "\n")
      file: t.Optional[t.IO[str]]:  (Default value = None)
      flush: bool:  (Default value = False)
    """
    import inspect
    from rich import print as rprint
    from pydantic import BaseModel
    from pipelime.cli.pretty_print import print_model_info
    from pipelime.sequences import SamplesSequence
    from pipelime.piper import PipelimeCommand

    for obj in objects:
        if isinstance(obj, (str, bytes)):
            if isinstance(obj, bytes):
                obj = obj.decode("utf-8")
            print_command_op_stage_info(obj)
        elif inspect.isclass(obj) and issubclass(obj, BaseModel):
            print_model_info(
                obj,
                show_class_path=(not issubclass(obj, SamplesSequence)),
                show_piper_port=issubclass(obj, PipelimeCommand),
            )
        else:
            sobj = (
                obj.__pl_pretty__()  # type: ignore
                if hasattr(obj, "__pl_pretty__")
                else str(obj)
            )
            rprint(sobj, sep=sep, end=end, file=file, flush=flush)


def create_stage_from_config(
    stage_name: str, stage_args: t.Union[t.Mapping[str, t.Any], t.Sequence, None]
) -> "SampleStage":  # type: ignore # noqa: E602,F821
    """Creates a stage from a name and arguments.

    Args:
        stage_name (str): the name of the stage, eg, `compose`, `remap`, etc.
        stage_args (t.Union[t.Mapping[str, t.Any], t.Sequence, None]): a mapping of the
            stage init arguments.

    Raises:
        ValueError: If the stage name is not a valid stage.

    Returns:
        SampleStage: The stage instance.
    """

    from pipelime.cli.utils import PipelimeSymbolsHelper
    from pipelime.stages import SampleStage

    stage_cls = PipelimeSymbolsHelper.get_stage(stage_name)
    if stage_cls is None or not issubclass(stage_cls[1], SampleStage):
        PipelimeSymbolsHelper.show_error_and_help(
            stage_name,
            should_be_cmd=False,
            should_be_op=False,
            should_be_stage=True,
        )
        raise ValueError(f"{stage_name} is not a pipelime stage.")
    stage_cls = stage_cls[1]
    if stage_args is None:
        return stage_cls()
    try:
        if isinstance(stage_args, (str, bytes)) or not isinstance(
            stage_args, (t.Sequence, t.Mapping)
        ):
            stage_args = [stage_args]
        return (
            stage_cls(**stage_args)
            if isinstance(stage_args, t.Mapping)
            else stage_cls(*stage_args)
        )
    except TypeError:
        # try to call without expanding args
        return stage_cls(stage_args)  # type: ignore


def get_pipelime_command_cls(
    cmd_name: str, interactive: bool = True
) -> t.Type["PipelimeCommand"]:
    from pipelime.piper.model import PipelimeCommand

    cmd_cls = PipelimeSymbolsHelper.get_command(cmd_name)
    if cmd_cls is None or not issubclass(cmd_cls[1], PipelimeCommand):
        if interactive:
            PipelimeSymbolsHelper.show_error_and_help(
                cmd_name, should_be_cmd=True, should_be_op=False, should_be_stage=False
            )
        raise ValueError(f"{cmd_name} is not a pipelime command.")
    return cmd_cls[1]


def get_pipelime_command(
    cmd: t.Union[t.Mapping[str, t.Optional[t.Mapping[str, t.Any]]], "PipelimeCommand"]
) -> "PipelimeCommand":
    from pipelime.piper.model import PipelimeCommand

    if isinstance(cmd, PipelimeCommand):
        return cmd

    cmd_name, cmd_args = next(iter(cmd.items()))
    cmd_cls = get_pipelime_command_cls(cmd_name)
    return cmd_cls() if cmd_args is None else cmd_cls(**cmd_args)


def time_to_str(nanosec: int) -> str:
    from datetime import timedelta

    if nanosec < 1000000000:
        millisec = int(nanosec // 1000000)
        nanosec -= millisec * 1000000
        microsec = int(nanosec // 1000)
        nanosec -= microsec * 1000
        return f"{millisec}ms {microsec}us {nanosec}ns"
    return str(timedelta(microseconds=nanosec / 1000))


def show_field_alias_valerr(e: ValidationError):
    def _replace_alias(val):
        if isinstance(val, str):
            for field in e.model.__fields__.values():  # type: ignore
                if (
                    field.model_config.allow_population_by_field_name
                    and field.has_alias
                    and field.alias == val
                ):
                    return f"{field.name} / {field.alias}"
        return val

    for err in e.errors():
        if "loc" in err:
            err["loc"] = tuple(_replace_alias(l) for l in err["loc"])



from typing import Dict
from pipelime.piper.progress.listener.base import ListenerCallback
from pipelime.piper.progress.model import OperationInfo, ProgressUpdate
from pipelime.piper.progress.tracker.base import TqdmTask


class TqdmBarsListenerCallback(ListenerCallback):
    """Listener callback based on tqdm progress bars."""

    def __init__(self) -> None:
        super().__init__()
        self._bars: Dict[OperationInfo, TqdmTask] = {}

    def on_start(self) -> None:
        from rich import print as rp
        from rich.align import Align
        from rich.panel import Panel
        from rich.style import Style
        from rich import box
        from rich.rule import Rule

        rp(
            Panel(
                Align("[red bold][on white]Piper Watcher[/][/]", align="center"),
                box=box.HEAVY,
            )
        )

    def on_update(self, prog: ProgressUpdate):
        if prog.op_info not in self._bars:
            self._bars[prog.op_info] = TqdmTask(
                total=prog.op_info.total,
                message="",
                bar_width=20,
                total_width=-1,
            )

            # adapt the width of the description to the longest one
            mlen = max(self._bars, key=lambda x: len(x.node) + len(x.message))
            mlen = len(mlen.node) + len(mlen.message) + 3
            for k, v in self._bars.items():
                v.set_message((f"[{k.node}] {k.message}").ljust(mlen))

        bar = self._bars[prog.op_info]

        # do not close on "finish", otherwise we won't be able to resize the bars
        if not prog.finished:
            bar.update(prog.progress)

    def on_stop(self):
        pass



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Mapping, Optional

from pipelime.piper.model import DAGModel


class DAGParser(ABC):
    @abstractmethod
    def parse_cfg(self, cfg: Mapping, params: Optional[Mapping] = None) -> DAGModel:
        """Parses the given configuration into a DAGModel.

        Args:
            cfg (Mapping): The input configuration as dictionary
                params (Optional[Mapping], optional): The parameters dictionary.
                Defaults to None.

        Returns:
            DAGModel: The parsed DAGModel
        """
        pass

    @abstractmethod
    def _read_file(
        self, path: Path, additional_args: Optional[Mapping] = None
    ) -> Mapping:
        """Reads the given file and returns its content as dictionary.

        Args:
            path (Path): The path to the file
            additional_args (Optional[Mapping], optional): Additional parameters that
                will overwrite the ones specified in the file. Defaults to None.

        Returns:
            Mapping: The file content as dictionary
        """
        pass

    def parse_file(
        self,
        cfg_file: Path,
        params_file: Optional[Path] = None,
        additional_args: Optional[Mapping] = None,
    ) -> DAGModel:
        """Parse the given configuration file into a DAGModel.

        Args:
            cfg_file (str): The input configuration file
                params_file (Optional[str], optional): The parameters file.
                Defaults to None.
            additional_args (Optional[Mapping], optional): Additional custom parameters
                that will overwrite the ones specified in params_file. Defaults to None.

        Returns:
            DAGModel: The parsed DAGModel
        """
        cfg = self._read_file(cfg_file)
        if params_file is not None:
            params = self._read_file(params_file, additional_args=additional_args)
        elif additional_args is not None:
            params = additional_args
        else:
            params = {}
        return self.parse_cfg(cfg, params)



from __future__ import annotations
import typing as t
import inspect
from pathlib import Path
import pydantic as pyd
import pydantic.generics as pydg
import numpy as np

from pipelime.items import Item

if t.TYPE_CHECKING:
    from pipelime.sequences import SamplesSequence
    from numpy.typing import ArrayLike


class NewPath(Path):
    @classmethod
    def __modify_schema__(cls, field_schema: t.Dict[str, t.Any]) -> None:
        field_schema.update(format="new-path")

    @classmethod
    def __get_validators__(cls):
        from pydantic.validators import path_validator

        yield path_validator
        yield cls.validate

    @classmethod
    def validate(cls, value: Path) -> Path:
        if value.exists():
            raise ValueError(f"Path {value} already exists")
        return value


class NumpyType(
    pyd.BaseModel,
    extra="forbid",
    copy_on_model_validation="none",
    arbitrary_types_allowed=True,
):
    """Numpy array type for stages, commands and any other pydantic model.
    Any argument accepted by `numpy.array()` is a valid value. Also, any mapping
    will be treated as keyword arguments for `numpy.array()`.

    Examples:
        Create a new NumpyType instance::

            npt = NumpyType.create(numpy.array([1,2,3]))
            npt = NumpyType.create([[1, 2, 3], [4, 5, 6]])
            npt = NumpyType.create(
                {
                    "object": [[1, 2, 3], [4, 5, 6]],
                    "dtype": "float32",
                    "order": "F",
                }
            )

        Access the numpy array::

            npt.value  # numpy array

        Serialize to dict or json::

            npt_dict = npt.dict()
            npt_json_str = npt.json()

        Get the object back from dict or json::

            npt_again = pydantic.parse_obj_as(NumpyType, npt_dict["__root__"])
            npt_again = pydantic.parse_raw_as(NumpyType, npt_json_str)

        Use this type within another model::

            class MyModel(pydantic.BaseModel):
                tensor: NumpyType = pydantic.Field(
                    default_factory=lambda: NumpyType.create([1, 2, 3])
                )

        Everything still works::

            mm = MyModel()
            mm = MyModel(tensor=np.array([1,2,3]))
            mm = MyModel.parse_obj({"tensor": [1, 2, 3]})
            mm = MyModel.parse_obj({"tensor": {"object": [1,2,3], "dtype": "float32"}})
            mm_again = pydantic.parse_obj_as(MyModel, mm.dict())
    """

    __root__: np.ndarray

    @classmethod
    def create(
        cls, value: t.Union[NumpyType, "ArrayLike", t.Mapping[str, t.Any]]
    ) -> NumpyType:
        return cls.validate(value)

    @property
    def value(self):
        return self.__root__

    def _iter(self, *args, **kwargs):
        for k, v in super()._iter(*args, **kwargs):
            # assert k == "__root__"
            # assert isinstance(v, np.ndarray)
            v_order = {} if v.flags["C_CONTIGUOUS"] else {"order": "F"}
            yield k, {"object": v.tolist(), "dtype": v.dtype.name, **v_order}

    def __str__(self) -> str:
        return str(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        try:
            return cls(
                __root__=(
                    np.array(**value)
                    if isinstance(value, t.Mapping)
                    else np.array(value)
                )
            )
        except Exception as e:
            raise ValueError(f"Invalid numpy input: {value}") from e


yaml_any_type = t.Union[
    None,
    pyd.StrictBool,
    pyd.StrictInt,
    pyd.StrictFloat,
    pyd.StrictStr,
    t.Mapping[str, t.Any],
    t.Sequence,
]


class YamlInput(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    """General yaml/json data (str, number, mapping, list...) optionally loaded from
    a yaml/json file, possibly with key path (format <filepath>[:<key>]).

    Examples:
        Create a new YamlInput from local data::

            yml = YamlInput.create(None)
            yml = YamlInput.create("some string")
            yml = YamlInput.create(4.2)
            yml = YamlInput.create([1, True, "string"])
            yml = YamlInput.create({"a": 1, "b": {"c": [1, 2, 3]}})

        Create a new YamlInput from a file::

            yml = YamlInput.create("path/to/file.yaml")
            yml = YamlInput.create("path/to/file.json")

        Create a new YamlInput from a pydash-like address within a file::

            yml = YamlInput.create("path/to/file.yaml:my.key.path")
            yml = YamlInput.create("path/to/file.json:my.key.path")

        Access the data::

            yml.value

        Serialize to dict or json::

            yml_dict = yml.dict()
            yml_json_str = yml.json()

        Get the object back from dict or json::

            yml_again = pydantic.parse_obj_as(YamlInput, yml_dict["__root__"])
            yml_again = pydantic.parse_raw_as(YamlInput, yml_json_str)

        Use this type within another model::

            class MyModel(pydantic.BaseModel):
                config: YamlInput = pydantic.Field(
                    default_factory=lambda: YamlInput.create([1, 2, 3])
                )

        Everything still works::

            mm = MyModel()
            mm = MyModel(config=[4, 5, 6])
            mm = MyModel.parse_obj({"config": [4, 5, 6]})
            mm_again = pydantic.parse_obj_as(MyModel, mm.dict())
    """

    __root__: yaml_any_type

    @classmethod
    def create(cls, value: t.Union[YamlInput, yaml_any_type]) -> YamlInput:
        return cls.validate(value)

    @property
    def value(self):
        return self.__root__

    def __str__(self) -> str:
        return str(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        if isinstance(value, (str, Path)):
            pval = Path(value)
            filepath, _, root_key = pval.name.partition(":")
            filepath = Path(pval.parent / filepath)
            if filepath.exists():
                import yaml
                import pydash as py_

                with filepath.open() as f:
                    value = yaml.safe_load(f)
                    if root_key:
                        value = py_.get(value, root_key, default=None)
            return cls(__root__=value)  # type: ignore
        if cls._check_any_type(value):
            return cls(__root__=value)
        raise ValueError(f"Invalid yaml data input: {value}")

    @classmethod
    def _check_any_type(cls, value):
        if isinstance(value, (str, int, float, bool, t.Sequence)) or value is None:
            return True
        if isinstance(value, t.Mapping):
            return all(isinstance(k, str) for k in value)


TRoot = t.TypeVar("TRoot")


class TypeDef(
    pydg.GenericModel,
    t.Generic[TRoot],
    extra="forbid",
    copy_on_model_validation="none",
    allow_mutation=False,
):
    """Generic type definition. It accepts both type names and string.
    You should derive from this class to define your own type definitions and,
    possibly, re-implement the `default_class_path` class method
    (NB: it must end with `.`). When a string is given, it can be a class path
    (the default class path can be omitted) or a `path/to/file.py:ClassName`.

    Examples:
        Create a new FooType definition::

            class FooType(TypeDef[Foo]):
                @classmethod
                def default_class_path(cls) -> str:
                    return "my_package.types."

        Now you can instantiate it::

            it = FooType.create(a.class.path.DerivedFromFoo)
            it = FooType.create("FooSubclassInMyPackage")

        Access the internal type and instantiate a new object::

            it = FooType.create(a.class.path.DerivedFromFoo)
            it.value  # DerivedFromFoo
            _ = it()  # new instance of DerivedFromFoo

        Serialize to dict or json::

            it_dict = it.dict()
            it_json_str = it.json()

        Get the object back from dict or json::

            it_again = pydantic.parse_obj_as(FooType, it_dict["__root__"])
            it_again = pydantic.parse_raw_as(FooType, it_json_str)

        Use this type within another model::

            class MyModel(pydantic.BaseModel):
                foo_type: FooType = pydantic.Field(
                    default_factory=lambda: FooType.create("DefaultFooSubclass")
                )

        Everything still works::

            mm = MyModel()
            mm = MyModel.parse_obj({"foo_type": a.class.path.DerivedFromFoo})
            mm = MyModel.parse_obj({"foo_type": "FooSubclassInMyPackage"})
            mm_again = pydantic.parse_obj_as(MyModel, mm.dict())
    """

    __root__: t.Type[TRoot]

    @classmethod
    def default_class_path(cls) -> str:
        return "__main__."

    @classmethod
    def wrapped_type(cls) -> t.Type[TRoot]:
        return t.get_args(cls.__fields__["__root__"].outer_type_)[0]

    @classmethod
    def create(cls, value: t.Union[TypeDef, t.Type[TRoot], str]) -> TypeDef:
        return cls.validate(value)

    @property
    def value(self) -> t.Type[TRoot]:
        return self.__root__

    def _iter(self, *args, **kwargs):
        for k, v in super()._iter(*args, **kwargs):
            # assert k == "__root__"
            # assert issubclass(v, self.wrapped_type())
            yield k, self._type_to_string(v)

    @classmethod
    def _type_to_string(cls, type_: t.Type[TRoot]) -> str:
        full_name = type_.__module__ + "." + type_.__qualname__
        redux_name = (
            full_name[len(cls.default_class_path()) :]
            if full_name.startswith(cls.default_class_path())
            else full_name
        )
        return full_name if "." in redux_name else redux_name

    @classmethod
    def _string_to_type(cls, type_str: str) -> t.Type[TRoot]:
        from pipelime.choixe.utils.imports import import_symbol

        type_str = type_str.strip("\"'")
        if "." not in type_str:
            type_str = cls.default_class_path() + type_str
        return import_symbol(type_str)

    def __call__(self, *args, **kwargs) -> TRoot:
        return self.__root__(*args, **kwargs)

    def __hash__(self) -> int:
        return hash(self.__root__)

    def __str__(self) -> str:
        return self._type_to_string(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value: t.Union[TypeDef, t.Type[TRoot], str]) -> TypeDef:
        import inspect

        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            value = cls._string_to_type(value)
        if inspect.isclass(value) and issubclass(value, cls.wrapped_type()):
            return cls(__root__=value)
        raise ValueError(f"Type `{value}` is not a subclass of `{cls.wrapped_type()}`")


class ItemType(TypeDef[Item]):
    """Item type definition. It accepts both type names and string.
    The default class path is `pipelime.items`.
    """

    @classmethod
    def default_class_path(cls) -> str:
        return "pipelime.items."


class CallableDef(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none", allow_mutation=False
):
    """Generic callable definition. It accepts functions and callable classes.
    You may derive from this class to re-implement the `default_class_path` class method
    (NB: it must end with `.`).

    Can be created from a symbol, an instance, a class/file path to a function or
    a mapping where the key is the class/file path to a class and the value is the
    list of __init__ arguments (mapping, sequence or single value).

    To ease the inspection of the callable, the methods `full_signature`, `args_type`,
    `return_type`, `has_var_positional` and `has_var_keyword` are provided.

    Examples:
        Create a new CallableDef::

            cdef = CallableDef.create(a.class.path.to.callable)
            cdef = CallableDef.create("CallableInMain")

        Access the internal value and call it::

            cdef = CallableDef.create(a.class.path.to.callable)
            cdef.value  # callable
            _ = cdef()  # call the callable

        Serialize to dict or json::

            cdef_dict = cdef.dict()
            cdef_json_str = cdef.json()

        Get the object back from dict or json::

            cdef_again = pydantic.parse_obj_as(CallableDef, cdef_dict["__root__"])
            cdef_again = pydantic.parse_raw_as(CallableDef, cdef_json_str)

        Use this type within another model::

            class MyModel(pydantic.BaseModel):
                fn: CallableDef = pydantic.Field(
                    default_factory=lambda: CallableDef.create("CallableInMain")
                )

        Everything still works::

            mm = MyModel()
            mm = MyModel.parse_obj({"fn": (a.class.path.to.callable})
            mm = MyModel.parse_obj({"fn": "CallableInMain"})
            mm_again = pydantic.parse_obj_as(MyModel, mm.dict())
    """

    __root__: t.Callable

    @classmethod
    def default_class_path(cls) -> str:
        return "__main__."

    @classmethod
    def create(cls, value: t.Union[CallableDef, t.Callable, str]) -> CallableDef:
        return cls.validate(value)

    @property
    def value(self) -> t.Callable:
        return self.__root__

    @property
    def full_signature(self) -> inspect.Signature:
        return inspect.signature(self.__root__)

    @property
    def args(self) -> t.Sequence[inspect.Parameter]:
        return list(self.full_signature.parameters.values())

    @property
    def args_type(self) -> t.Sequence[t.Optional[t.Type]]:
        return [
            None if p.annotation is inspect.Signature.empty else p.annotation
            for p in self.full_signature.parameters.values()
        ]

    @property
    def has_var_positional(self) -> bool:
        return any(
            p.kind is p.VAR_POSITIONAL for p in self.full_signature.parameters.values()
        )

    @property
    def has_var_keyword(self) -> bool:
        return any(
            p.kind is p.VAR_KEYWORD for p in self.full_signature.parameters.values()
        )

    @property
    def return_type(self) -> t.Optional[t.Type]:
        rt = self.full_signature.return_annotation
        return None if rt is inspect.Signature.empty else rt

    def _iter(self, *args, **kwargs):
        for k, v in super()._iter(*args, **kwargs):
            # assert k == "__root__"
            yield k, self._callable_to_string(v)

    @classmethod
    def _callable_to_string(cls, clb: t.Callable) -> str:
        full_name = clb.__module__ + "." + clb.__qualname__
        redux_name = (
            full_name[len(cls.default_class_path()) :]
            if full_name.startswith(cls.default_class_path())
            else full_name
        )
        return full_name if "." in redux_name else redux_name

    @classmethod
    def _string_to_callable(cls, clb_str: str) -> t.Callable:
        from pipelime.choixe.utils.imports import import_symbol

        clb_str = clb_str.strip("\"'")
        if "." not in clb_str:
            clb_str = cls.default_class_path() + clb_str
        return import_symbol(clb_str)

    def __call__(self, *args, **kwargs):
        return self.__root__(*args, **kwargs)

    def __hash__(self) -> int:
        return hash(self.__root__)

    def __str__(self) -> str:
        return self._callable_to_string(self.__root__)

    def __repr__(self) -> str:
        return self.__piper_repr__()

    def __piper_repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        value: t.Union[
            CallableDef, t.Callable, str, t.Mapping[t.Union[str, t.Callable], t.Any]
        ],
    ):
        if isinstance(value, cls):
            return value
        try:
            if isinstance(value, str):
                value = cls._string_to_callable(value)
            elif isinstance(value, t.Mapping):
                clb_path, clb_args = next(iter(value.items()))
                clb = (
                    cls._string_to_callable(clb_path)
                    if isinstance(clb_path, str)
                    else clb_path
                )

                if not isinstance(clb, t.Callable):
                    raise ValueError(f"Invalid callable: {clb_path}")

                if isinstance(clb_args, t.Mapping):
                    value = clb(**clb_args)
                elif isinstance(clb_args, t.Sequence) and not isinstance(
                    clb_args, (str, bytes)
                ):
                    value = clb(*clb_args)
                else:
                    value = clb(clb_args)
        except Exception as e:
            raise ValueError(f"Invalid callable: {value}") from e

        if isinstance(value, t.Callable):
            return cls(__root__=value)
        raise ValueError(f"Invalid callable: {value}")


# This is defined here to make it picklable
def _identity_fn_helper(x):
    return x


class ItemValidationModel(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """Item schema validation."""

    class_path: ItemType = pyd.Field(
        ...,
        description=(
            "The item class path. The default package `pipelime.item` can be omitted"
        ),
    )
    is_optional: bool = pyd.Field(
        True, description="Whether the item is required or optional."
    )
    is_shared: bool = pyd.Field(False, description="Whether the item is shared or not.")
    validator_: t.Optional[str] = pyd.Field(
        None,
        description=(
            "A class path to a callable accepting the item value and either returning "
            "a validated value or raising an exception in case of error."
        ),
        alias="validator",
    )

    _validator_callable = pyd.PrivateAttr()

    def __init__(self, **data):
        from pipelime.choixe.utils.imports import import_symbol

        super().__init__(**data)
        self._validator_callable = (
            import_symbol(self.validator_) if self.validator_ else _identity_fn_helper
        )

    def make_field(self, key_name: str):
        return (
            self.class_path.value,
            pyd.Field(default_factory=self.class_path.value, alias=key_name)
            if self.is_optional
            else pyd.Field(..., alias=key_name),
        )

    def make_validator_method(self, field_name: str):
        import uuid

        # we need random names and dynamic function creation
        # to avoid reusing the same function name for validators
        # (yes, pydantic is really pedantic...)
        rnd_name = uuid.uuid1().hex

        _validator_wrapper = (
            "def validate_{}_fn(cls, v):\n".format(rnd_name)
            + "    if v.is_shared != {}:\n".format(self.is_shared)
            + "        raise ValueError(\n"
            + "            'Item must{}be shared.'\n".format(
                " not " if not self.is_shared else " "
            )
            + "        )\n"
            + "    return user_validator_{}(v)\n".format(rnd_name)
        )

        local_scope = {
            **globals(),
            f"user_validator_{rnd_name}": self._validator_callable,
        }
        exec(_validator_wrapper, local_scope)
        fn_helper = local_scope[f"validate_{rnd_name}_fn"]
        return pyd.validator(field_name)(fn_helper)


class SampleValidationInterface(
    pyd.BaseModel, extra="forbid", copy_on_model_validation="none"
):
    """Sample schema validation."""

    sample_schema: t.Union[
        t.Type[pyd.BaseModel], str, t.Mapping[str, ItemValidationModel]
    ] = pyd.Field(
        ...,
        description=(
            "The sample schema to validate, ie, a mapping from sample keys to expected "
            "item types.\nThe schema can be a pydantic Model or a class path to a "
            "pydantic model to import, where fields' names are the sample keys, "
            "while fields' values are the item types. Otherwise, an explicit "
            "`key-name: ItemValidationModel` mapping must be provided."
        ),
    )
    ignore_extra_keys: bool = pyd.Field(
        True,
        description=(
            "When `sample_schema` is an explicit mapping, if `ignore_extra_keys` is "
            "True, unexpected keys are ignored. Otherwise an error is raised."
        ),
    )
    lazy: bool = pyd.Field(
        True, description="If True, samples will be validated only when accessed."
    )
    max_samples: int = pyd.Field(
        1,
        description=(
            "When the validation is NOT lazy, "
            "only the slice `[0:max_samples]` is checked. "
            "Set to 0 to check all the samples."
        ),
    )

    _schema_model: t.Optional[t.Type[pyd.BaseModel]] = pyd.PrivateAttr(None)

    def _import_schema(self, schema_path: str):
        from pipelime.choixe.utils.imports import import_symbol

        imported_schema = import_symbol(schema_path)
        return imported_schema

    def _make_schema(self, schema_def: t.Mapping[str, ItemValidationModel]):
        class Config(pyd.BaseConfig):
            arbitrary_types_allowed = True
            extra = pyd.Extra.ignore if self.ignore_extra_keys else pyd.Extra.forbid

        def _safe_name(k):
            return f"{k}___"

        _item_map = {_safe_name(k): v.make_field(k) for k, v in schema_def.items()}
        _validators = {
            f"validate_{k}": v.make_validator_method(_safe_name(k))
            for k, v in schema_def.items()
        }

        return pyd.create_model(
            "SampleSchema",
            __config__=Config,
            __validators__=_validators,
            **_item_map,
        )

    @property
    def schema_model(self) -> t.Type[pyd.BaseModel]:
        sm = self._schema_model
        if sm is None:
            if isinstance(self.sample_schema, str):
                sm = self._import_schema(self.sample_schema)
            elif isinstance(self.sample_schema, t.Mapping):
                sm = self._make_schema(self.sample_schema)
            else:
                sm = self.sample_schema

            if not issubclass(sm, pyd.BaseModel):
                raise ValueError(f"`{self.sample_schema}` is not a pydantic model.")

            # cache the model for later use
            if self.lazy:
                self._schema_model = sm
        return sm

    def append_validator(self, sequence: "SamplesSequence") -> "SamplesSequence":
        return sequence.validate_samples(sample_schema=self)

    def as_pipe(self):
        return {"validate_samples": {"sample_schema": self.dict(by_alias=True)}}



import inspect
import typing as t

from pydantic import BaseModel
from rich import box
from rich.markup import escape
from rich import get_console
from rich import print as rprint
from rich.pretty import Pretty
from rich.table import Table, Column
from rich.text import Text
from rich.style import Style
from rich.panel import Panel

if t.TYPE_CHECKING:
    from pipelime.cli.utils import ActionInfo


def _input_icon():
    return "📥"


def _output_icon():
    return "📦"


def _parameter_icon():
    return "📐"


def _short_line():
    return "━━━━━"


def print_with_style(
    val,
    style: t.Optional[str] = None,
    *,
    pretty: bool = False,
    end: str = "\n",
    indent_guides: bool = True,
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style=style,
        end=end,
    )


def print_debug(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="italic grey50",
        end=end,
    )


def print_info(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="cyan",
        end=end,
    )


def print_warning(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        "[bold blink]WARNING:[/bold blink]",
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="orange1",
        end=end,
    )


def print_error(
    val, *, pretty: bool = False, end: str = "\n", indent_guides: bool = True
):
    get_console().print(
        "[bold blink]ERROR:[/bold blink]",
        Pretty(val, indent_guides=indent_guides, expand_all=True) if pretty else val,
        style="dark_red on white",
        end=end,
    )


def show_spinning_status(text: str):
    """Returns a context manager."""
    return get_console().status(text)


def get_model_title(model_cls: t.Type[BaseModel]) -> str:
    if model_cls.__config__.title:
        return model_cls.__config__.title
    return model_cls.__name__


def get_model_classpath(model_cls: t.Type[BaseModel]) -> str:
    if hasattr(model_cls, "_classpath") and model_cls._classpath:  # type: ignore
        return model_cls._classpath  # type: ignore
    return f"{model_cls.__module__}.{model_cls.__name__}"


def print_model_field_values(
    model_fields: t.Mapping,
    port_values: t.Mapping[str, t.Any],
    icon: str = "",
):
    for k, v in port_values.items():
        rprint(f"\n{icon if icon else '***'} {k}:")
        # Ports might be virtual, as in ShellCommand, so they might not be in the model
        if k in model_fields and model_fields[k].field_info.description:
            rprint(
                f"[italic grey50]{escape(model_fields[k].field_info.description)}[/]"
            )
        rprint(
            "[green]"
            + escape(str(v) if isinstance(v, (bytes, str)) else repr(v))
            + "[/]"
        )


def print_command_inputs(command: "PipelimeCommand"):  # type: ignore # noqa: E602,F821
    print_model_field_values(command.__fields__, command.get_inputs(), _input_icon())


def print_command_outputs(command: "PipelimeCommand"):  # type: ignore # noqa: E602,F821
    print_model_field_values(command.__fields__, command.get_outputs(), _output_icon())


def print_actions_short_help(*actions_info: "ActionInfo", show_class_path: bool = True):
    grid = Table.grid(
        *([Column(overflow="fold") for _ in range(2 + int(show_class_path))]),
        padding=(0, 1),
    )
    for a in actions_info:
        col_vals = [escape(a.name)]
        if show_class_path:
            col_vals.append(f"[italic grey50]{escape(a.classpath)}[/]")
        col_vals.append(escape(a.description))
        grid.add_row(*col_vals)
    rprint(grid)


def print_models_short_help(
    *model_cls: t.Type[BaseModel], show_class_path: bool = True
):
    grid = Table.grid(
        *([Column(overflow="fold") for _ in range(2 + int(show_class_path))]),
        padding=(0, 1),
    )
    for m in model_cls:
        col_vals = [escape(get_model_title(m))]
        if show_class_path:
            col_vals.append(f"[italic grey50]{escape(get_model_classpath(m))}[/]")
        col_vals.append(escape(inspect.getdoc(m) or ""))
        grid.add_row(*col_vals)
    rprint(grid)


def print_model_info(
    model_cls: t.Type[BaseModel],
    *,
    indent_offs: int = 2,
    show_class_path: bool = True,
    show_piper_port: bool = True,
    show_description: bool = True,
    recursive: bool = True,
):
    model_docs = (inspect.getdoc(model_cls) or "") if show_description else ""

    # * Fields, [Description], Type, [Piper Port], Default
    cols = [Column("Fields", overflow="fold")]
    if show_description:
        cols.append(Column("Description", overflow="fold"))
    cols.append(Column("Type", overflow="fold"))
    if show_piper_port:
        cols.append(Column("Piper Port", overflow="fold"))
    cols.append(Column("Default", overflow="fold"))

    grid = Table(
        *cols,
        box=box.SIMPLE_HEAVY,
        title=(
            f"[#5fafff]{_get_signature(model_cls)}[/]"
            + (f"\n\n[italic grey82]{escape(model_docs)}[/]" if model_docs else "")
        ),
        title_style="on #293a05",
        title_justify="left",
        expand=True,
    )

    _iterate_model_fields(
        model_cls=model_cls,
        grid=grid,
        indent=0,
        indent_offs=indent_offs,
        show_piper_port=show_piper_port,
        show_description=show_description,
        recursive=recursive,
        add_blank_row=True,
    )

    grid = Panel(
        grid,
        title=f"[dark_orange bold][on #293a05]{escape(get_model_title(model_cls))}[/]",
        subtitle=(
            f"[#5fafff][on #293a05]{escape(get_model_classpath(model_cls))}[/]"
            if show_class_path
            else None
        ),
    )

    rprint(grid)


def _field_row(
    grid: Table,
    field,
    indent: int,
    indent_offs: int,
    show_piper_port: bool,
    show_description: bool,
    recursive: bool,
):
    is_model = _is_model(field.outer_type_) and not inspect.isabstract(
        field.outer_type_
    )

    if show_description:
        # NB: docs should not come from the inner __root__ type
        if field.field_info.description:
            field_docs = field.field_info.description
        elif hasattr(field.outer_type_, "__doc__") and field.outer_type_.__doc__:
            field_docs = str(inspect.getdoc(field.outer_type_))
        else:
            field_docs = ""

        # field_docs = " ".join(field_docs.split())

    has_root_item = ("__root__" in field.outer_type_.__fields__) if is_model else False
    field_outer_type = (
        field.outer_type_.__fields__["__root__"].outer_type_
        if has_root_item
        else field.outer_type_
    )

    if show_piper_port:
        from pipelime.piper import PiperPortType

        fport = str(
            field.field_info.extra.get("piper_port", PiperPortType.PARAMETER).value
        ).upper()

        if fport == PiperPortType.INPUT.value.upper():
            fport = f"{_input_icon()} [yellow]{fport}[/]"
        elif fport == PiperPortType.OUTPUT.value.upper():
            fport = f"{_output_icon()} [cyan]{fport}[/]"
        else:
            fport = f"{_parameter_icon()} {fport}"

    # Field name & alias
    line = [
        (" " * indent)
        + ("[bold dark_orange]" if indent == 0 else "")
        + (
            f"{escape(field.name)} / "
            if field.model_config.allow_population_by_field_name and field.has_alias
            else ""
        )
        + f"{escape(field.alias)}"
        + ("[/]" if indent == 0 else "")
    ]

    # Description
    if show_description:
        line.append("▶ " + escape(field_docs))  # type: ignore

    # Type
    line.append(
        ""
        if is_model and not has_root_item and recursive
        else _human_readable_type(field_outer_type).replace("[", r"\[")  # noqa: W605
    )

    # Piper port
    if show_piper_port:
        line.append(fport)  # type: ignore

    # Default value
    line.append("[red]✗[/]" if field.required else f"[green]{field.get_default()}[/]")

    grid.add_row(*line)

    if recursive:
        if is_model and not has_root_item:
            _iterate_model_fields(
                model_cls=field_outer_type,
                grid=grid,
                indent=indent + indent_offs,
                indent_offs=indent_offs,
                show_piper_port=show_piper_port,
                show_description=show_description,
                recursive=recursive,
                add_blank_row=False,
            )
        else:
            inner_types = _get_inner_args(field_outer_type)
            last_types = inner_types
            while last_types:
                last_types = _get_inner_args(*last_types)
                inner_types |= last_types
            inner_types = {
                arg for arg in inner_types if _is_model(arg) and arg is not BaseModel
            }

            for arg in inner_types:
                grid.add_row(
                    (" " * indent)
                    + f"[grey50]{_short_line()} {arg.__name__}[/]"  # type:ignore
                )
                _iterate_model_fields(
                    model_cls=arg,
                    grid=grid,
                    indent=indent + indent_offs,
                    indent_offs=indent_offs,
                    show_piper_port=show_piper_port,
                    show_description=show_description,
                    recursive=recursive,
                    add_blank_row=False,
                )


def _get_signature(model_cls: t.Type[BaseModel]) -> str:
    class RichParameter(inspect.Parameter):
        def __init__(self, *, name, kind, default, annotation):
            self._name = name
            self._kind = kind
            self._default = default
            self._annotation = annotation

        def __str__(self):
            kind = self.kind
            formatted = self._name

            color = "dark_orange"
            if kind == inspect.Parameter.VAR_POSITIONAL:
                formatted = "*" + formatted
            elif kind == inspect.Parameter.VAR_KEYWORD:
                formatted = "**<any-extra-data>"
                color = "indian_red"

            formatted = f"[bold {color}]{escape(formatted)}[/]"

            # Add annotation and default value
            if self._annotation is not inspect._empty:
                formatted = "{}: {}".format(
                    formatted,
                    escape(
                        inspect.formatannotation(self._annotation).replace(
                            "NoneType", "None"
                        )
                    ),
                )

            if self._default is not inspect._empty:
                default_str = escape(repr(self._default))
                if self._annotation is not inspect._empty:
                    formatted = "{} = {}".format(formatted, default_str)
                else:
                    formatted = "{}={}".format(formatted, default_str)

            return f"\n  {formatted}"

    fullname = {mfield.alias: mfield.name for mfield in model_cls.__fields__.values()}
    excluded = [
        mfield.alias
        for mfield in model_cls.__fields__.values()
        if mfield.field_info.exclude
    ]

    sig = inspect.signature(model_cls)
    sig = sig.replace(
        parameters=[
            RichParameter(
                name=fullname.get(p.name, p.name),
                kind=p.kind,
                default=p.default,
                annotation=p.annotation,
            )
            for p in sig.parameters.values()
            if p.name not in excluded
        ],
        return_annotation=inspect.Signature.empty,
    )

    sig = (
        str(sig)
        .replace("/,", "\n  /,")
        .replace("*,", "\n  *,")
        .replace("/)", "\n  /)")
        .replace("*)", "\n  *)")
    )
    sig = sig[:-1] + "\n)"
    return sig


def _is_model(type_):
    return inspect.isclass(type_) and issubclass(type_, BaseModel)


def _human_readable_type(field_outer_type):
    from enum import Enum
    from pydantic.typing import (
        typing_base,  # noqa: F401  # type: ignore
        WithArgsTypes,
        is_union,
        get_origin,
        get_args,
    )

    v = field_outer_type
    if (
        not isinstance(v, typing_base)
        and not isinstance(v, WithArgsTypes)
        and not isinstance(v, type)
    ):
        v = v.__class__

    v_orig = get_origin(v)

    if is_union(v_orig):
        return " | ".join(map(_human_readable_type, get_args(v)))
    if inspect.isclass(v_orig):
        if issubclass(dict, v_orig):
            return "{" + ": ".join(map(_human_readable_type, get_args(v))) + "}"
        if issubclass(list, v_orig):
            return "[" + ", ".join(map(_human_readable_type, get_args(v))) + ", ...]"
        if issubclass(tuple, v_orig):
            return "(" + ", ".join(map(_human_readable_type, get_args(v))) + ")"

    if inspect.isclass(v) and issubclass(v, Enum):
        v = v.__name__ + "{" + ", ".join(e.name.lower() for e in v) + "}"
    elif isinstance(v, WithArgsTypes):
        # Generic alias are constructs like `list[int]`
        v = str(v).replace("typing.", "")
    else:
        try:
            v = v.__name__
        except AttributeError:
            # happens with typing objects
            v = str(v).replace("typing.", "")

    return v.replace("NoneType", "None")


def _recursive_args_flattening(arg):
    if isinstance(arg, t.Mapping):
        return {
            a
            for k, v in arg.items()
            for u in (_recursive_args_flattening(k), _recursive_args_flattening(v))
            for a in u
        }
    if isinstance(arg, t.Collection) and not isinstance(arg, (str, bytes)):
        return {a for v in arg for a in _recursive_args_flattening(v)}
    return {arg}


def _get_inner_args(*type_):
    return {
        a
        for t_ in type_
        for arg in t.get_args(t_)
        for a in _recursive_args_flattening(arg)
    }


def _iterate_model_fields(
    model_cls,
    grid,
    indent,
    indent_offs,
    show_piper_port,
    show_description,
    recursive,
    add_blank_row,
):
    no_data = True
    for field in model_cls.__fields__.values():  # type: ignore
        if not field.field_info.exclude:
            no_data = False
            _field_row(
                grid,
                field,
                indent=indent,
                indent_offs=indent_offs,
                show_piper_port=show_piper_port,
                show_description=show_description,
                recursive=recursive,
            )
            if add_blank_row:
                grid.add_row()
    if no_data:
        grid.add_row((" " * indent) + "[grey50 italic]([strike]no parameters[/])[/]")



from abc import ABC, abstractmethod
from itertools import count
from typing import Iterable, Optional, Sequence, Union

from pipelime.piper.progress.model import OperationInfo, ProgressUpdate


class TrackCallback(ABC):
    """A custom callback to track the progress of a running task"""

    @abstractmethod
    def update(self, prog: ProgressUpdate) -> None:
        """What to do when the operation advances

        Args:
            prog (ProgressUpdate): The progress update object.
        """


class TrackedTask(ABC):
    """Context manager to track a single task."""

    def __init__(self):
        self._progress = 0

    @property
    def progress(self) -> int:
        return self._progress

    def advance(self, advance: int = 1):
        """Advance the task by a custom amount of steps."""
        self._do_update(self._progress + advance)

    def update(self, progress: int):
        """Set the progress of the task to a custom value."""
        self._do_update(progress)

    def finish(self):
        """Complete the task."""
        self._do_update(self._progress, finished=True)

    def restart(self):
        """Restart the task."""
        self._do_update(0)

    def track(self, iterable: Iterable):
        """Track a generic iterable sequence"""
        for x in iterable:
            yield x
            self.advance()

    def _do_update(self, progress: int, finished: bool = False):
        self._progress = progress
        self.on_update(finished)

    @abstractmethod
    def on_update(self, finished: bool = False):
        """What to do when the task is updated"""

    def __enter__(self):
        self.restart()
        return self

    def __exit__(self, exc_type, exc, exc_tb):
        self.finish()


class PiperTask(TrackedTask):
    def __init__(self, op_info: OperationInfo, callbacks: Sequence[TrackCallback]):
        super().__init__()
        self._op_info = op_info
        self._callbacks = callbacks

    def on_update(self, finished: bool = False):
        prog = ProgressUpdate(
            op_info=self._op_info, progress=self.progress, finished=finished
        )
        for cb in self._callbacks:
            cb.update(prog)


class TqdmTask(TrackedTask):
    @staticmethod
    def default_bar(
        iterable=None,
        *,
        total=None,
        message=None,
        bar_width=None,
        ncols=None,
        position=None,
    ):
        from tqdm import tqdm

        bar = tqdm(
            iterable,
            total=len(iterable) if total is None else total,  # type: ignore
            colour="#4CAE4F",
            position=position,
            ncols=88 if ncols is None else (None if ncols <= 0 else ncols),
            bar_format=(
                None
                if bar_width is None
                else "{desc} {percentage:3.0f}%|{bar:" + str(bar_width) + "}{r_bar}"
            ),
        )
        TqdmTask._set_description(bar, message)
        return bar

    @staticmethod
    def _set_description(bar, message: Optional[str]):
        bar.set_description_str("🍋 " + message if message else "🍋")

    def __init__(
        self,
        total: int,
        message: str,
        bar_width: Optional[int] = None,
        total_width: Optional[int] = None,
        position: Optional[int] = None,
    ):
        super().__init__()
        self._bar = None
        self._total = total
        self._message = message
        self._bar_width = bar_width
        self._total_width = total_width
        self._position = position

    def set_message(self, message: str):
        self._message = message
        if self._bar:
            self._set_description(self._bar, self._message)

    @property
    def bar(self):
        # create and show the bar upon first access
        if not self._bar:
            self._bar = self.default_bar(
                total=self._total,
                message=self._message,
                bar_width=self._bar_width,
                ncols=self._total_width,
                position=self._position,
            )
        return self._bar

    def on_update(self, finished: bool = False):
        bar = self.bar
        if finished:
            bar.close()
        elif bar.n < self.progress:
            bar.update(self.progress - bar.n)
        elif bar.n > self.progress:
            bar.reset()
            bar.update(self.progress)


class Tracker:
    """Tracker for running operations"""

    def __init__(self, token: str, node: str, *callbacks: TrackCallback) -> None:
        """Constructor for a `Tracker`

        Args:
            token (str): The piper token
            node (str): The piper node name
            callbacks (TrackCallback, optional): The callbacks to use.
        """
        self._counter = count()
        self._callbacks = callbacks
        self._token = token
        self._node = node

    def track(
        self,
        seq: Union[Iterable, Sequence],
        size: Optional[int] = None,
        message: str = "",
    ) -> Iterable:
        """Track a generic iterable sequence"""

        with self.create_task(
            total=len(seq) if size is None else size, message=message  # type: ignore
        ) as t:
            yield from t.track(seq)

    def create_task(self, total: int, message: str = ""):
        """Explicit task creation"""

        op_info = OperationInfo(
            token=self._token,
            node=self._node,
            chunk=next(self._counter),
            total=total,  # type: ignore
            message=message,
        )
        return PiperTask(op_info, self._callbacks)



from pathlib import Path
from typing import Mapping, Optional

from pipelime.choixe import XConfig
from pipelime.piper.model import DAGModel
from pipelime.piper.parsers.base import DAGParser


class ChoixeDAGParser(DAGParser):
    def parse_cfg(self, cfg: Mapping, params: Optional[Mapping] = None) -> DAGModel:
        cfg = XConfig(cfg)
        context = XConfig(params)

        cfg = cfg.process(context)
        assert cfg.inspect().processed, "The configuration is still not processed"

        return DAGModel.parse_obj(cfg.to_dict())

    def _read_file(
        self, path: Path, additional_args: Optional[Mapping] = None
    ) -> Mapping:
        data = XConfig.from_file(path)
        additional_args = {} if additional_args is None else additional_args
        data.deep_update(additional_args)  # type: ignore
        return data



from typing import Sequence

from pipelime.piper.parsers.base import DAGParser
from pipelime.piper.parsers.choixe_parser import ChoixeDAGParser


class DAGParserFactory:
    DEFAULT_PARSER = "ChoixeDAGParser"

    _parsers_map = {
        "ChoixeDAGParser": ChoixeDAGParser,
    }

    @classmethod
    def available_parsers(cls) -> Sequence[str]:
        return list(DAGParserFactory._parsers_map.keys())

    @classmethod
    def get_parser(cls, parser_name: str = DEFAULT_PARSER, **kwargs) -> DAGParser:
        return cls._parsers_map[parser_name](**kwargs)  # type: ignore



from abc import ABC, abstractmethod
from typing import Optional

from loguru import logger

from pipelime.piper.graph import DAGNodesGraph, GraphNodeOperation
from pipelime.piper.progress.listener.base import Listener
from pipelime.piper.progress.listener.factory import (
    ListenerCallbackFactory,
    ProgressReceiverFactory,
)
from pipelime.piper.progress.tracker.base import TrackedTask


class NodesGraphExecutor(ABC):
    """An NodesGraphExecutor is an abstract class that should
    implement the execution of a NodesGraph made of Nodes.
    """

    def __init__(self, node_prefix: str) -> None:
        super().__init__()
        self._node_prefix = node_prefix
        self._task = None

    @property
    def task(self) -> Optional[TrackedTask]:
        return self._task

    @task.setter
    def task(self, task: Optional[TrackedTask]):
        self._task = task

    def _set_piper_info(self, node: GraphNodeOperation, token: str):
        node.command.set_piper_info(token=token, node=self._node_prefix + node.name)

    @abstractmethod
    def exec(self, graph: DAGNodesGraph, token: str = "") -> bool:
        """Executes the given NodesGraph.

        Args:
            graph (DAGNodesGraph): input DAGNodesGraph
            token (str, optional): execution token shared among nodes. Defaults to "".

        Returns:
            bool: TRUE if the execution was successful, FALSE otherwise
        """
        raise NotImplementedError()


class WatcherNodesGraphExecutor(NodesGraphExecutor):
    """Decorator for `NodesGraphExecutor` that wraps an existing executor.

    The `WatcherNodesGraphExecutor` disables all logs from the wrapped executor and
    lets a `Watcher` print live updates.

    This wrapper can be used with any subtype of `NodesGraphExecutor`::

        # Different types of executor
        executor1 = NaiveNodesGraphExecutor(token)
        executor2 = AdvancedNodesGraphExecutor(token)
        executor3 = SuperCoolNodesGraphExecutor(token)

        # All of them can be transformed into a `WatcherNodesGraphExecutor`
        # regardless of their type.
        executor1 = WatcherNodesGraphExecutor(executor1)
        executor2 = WatcherNodesGraphExecutor(executor2)
        executor3 = WatcherNodesGraphExecutor(executor3)
    """

    def __init__(
        self, executor: NodesGraphExecutor, listener_clbk: Optional[str] = None
    ) -> None:
        self._executor = executor
        self._listener_clbk = listener_clbk

    def exec(self, graph: DAGNodesGraph, token: str = "") -> bool:
        res = False
        logger.disable(self._executor.__module__)
        receiver = ProgressReceiverFactory.get_receiver(token)
        callback = ListenerCallbackFactory.get_callback(
            self._listener_clbk or ListenerCallbackFactory.DEFAULT_CALLBACK_TYPE
        )
        listener = Listener(receiver, callback)
        try:
            listener.start()
            res = self._executor.exec(graph, token)
        finally:
            listener.stop()
            logger.enable(self._executor.__module__)
        return res



import typing as t

import pydantic as pyd

from pipelime.stages import SampleStage
from pipelime.items import Item

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class ItemInfo(pyd.BaseModel):
    """Item infos estracted from samples."""

    item_type: t.Type[Item] = pyd.Field(..., description="The item type.")
    is_shared: bool = pyd.Field(..., description="Whether the item is shared.")
    count_: int = pyd.Field(
        1, alias="count", description="The number of samples owning this item."
    )


class StageItemInfo(SampleStage, title="item-info"):
    """Collects item infos from samples.
    WARNING: this stage CANNOT be combined with MULTIPROCESSING.
    """

    _items_info: t.MutableMapping[str, ItemInfo] = pyd.PrivateAttr(default_factory=dict)

    @property
    def items_info(self):
        return self._items_info

    def __call__(self, x: "Sample") -> "Sample":
        for k, v in x.items():
            class_name = (
                v.__class__.__name__
                if v.__class__.__module__.startswith("pipelime.items")
                else f"{v.__class__.__module__}.{v.__class__.__name__}"
            )
            if k in self._items_info:
                if self._items_info[k].item_type != v.__class__:
                    raise ValueError(
                        f"Key {k} has multiple types: "
                        f"{self._items_info[k].item_type} and {v.__class__}."
                    )
                if self._items_info[k].is_shared != v.is_shared:
                    raise ValueError(f"Key {k} is not always shared or not shared.")
                self._items_info[k].count_ += 1
            else:
                self._items_info[k] = ItemInfo(
                    item_type=v.__class__, is_shared=v.is_shared
                )  # type: ignore
        return x



import time
from datetime import datetime, timedelta
from threading import Thread
from typing import Dict

import numpy as np
from rich import box
from rich.live import Live
from rich.table import Table

from pipelime.piper.progress.listener.base import ListenerCallback
from pipelime.piper.progress.model import OperationInfo, ProgressUpdate
from pipelime.piper.progress.estimator.base import Estimator
from pipelime.piper.progress.estimator.factory import EstimatorFactory


class RichTableListenerCallback(ListenerCallback):
    """A callback for the listener that displays a rich table"""

    def __init__(self) -> None:
        super().__init__()
        self._printer_thread = None
        self.on_stop()

    def on_start(self) -> None:
        self._stop_flag = False
        self._printer_thread = Thread(target=self._print_loop)
        self._printer_thread.start()

    def on_update(self, prog: ProgressUpdate) -> None:
        op = prog.op_info
        if op not in self._estimators:
            estimator = EstimatorFactory.get_estimator()
            estimator.reset(op.total)
            self._estimators[op] = estimator
            self._progress_map[op] = prog

        advance = prog.progress - self._progress_map[op].progress
        if advance > 0:
            self._estimators[op].tick(advance)

        self._progress_map[op] = prog

    def _percentage_string(self, v: float):
        colors = ["red", "yellow", "green"]
        ths = np.linspace(0.0, 1.0, len(colors) + 1)[:-1]

        color = colors[0]
        for th, maybe_color in zip(ths, colors):
            if v > th:
                color = maybe_color

        return f"[{color}]{v*100:.1f}%[/{color}]"

    def _generate_table(self) -> Table:
        title = "Piper Watcher"
        table = Table(
            "Node",
            "Message",
            "Progress",
            "Start Time",
            "Elapsed",
            "Speed",
            "ETA",
            box=box.SIMPLE_HEAVY,
            title=f"[bold red]{title}[/]",
            title_style="on white",
        )
        for op in self._progress_map.keys():
            prog = self._progress_map[op]
            est = self._estimators[op]
            table.add_row(
                op.node,
                op.message,
                self._percentage_string(
                    prog.progress / op.total if op.total != 0 else 1.0
                ),
                datetime.fromtimestamp(est.start_time).strftime("%Y-%m-%d %H:%M:%S"),
                str(timedelta(seconds=round(est.elapsed))),
                "N.A." if est.speed < 0 else str(round(est.speed, 2)),
                "N.A." if est.eta < 0 else str(timedelta(seconds=round(est.eta))),
            )

        return table

    def _print_loop(self) -> None:
        with Live(self._generate_table(), refresh_per_second=4) as live:
            while not self._stop_flag:
                time.sleep(0.1)
                live.update(self._generate_table())

    def on_stop(self) -> None:
        self._stop_flag = True
        if self._printer_thread is not None:
            self._printer_thread.join(5.0)
        self._printer_thread = None
        self._progress_map: Dict[OperationInfo, ProgressUpdate] = {}
        self._estimators: Dict[OperationInfo, Estimator] = {}



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Sequence

import numpy as np

from pipelime.piper.graph import DAGNodesGraph


class NodesGraphDrawer(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def exportable_formats(self) -> Sequence[str]:
        raise NotImplementedError()

    @abstractmethod
    def draw(self, graph: DAGNodesGraph, **kwargs) -> np.ndarray:
        raise NotImplementedError()

    @abstractmethod
    def representation(self, graph: DAGNodesGraph) -> str:
        raise NotImplementedError()

    @abstractmethod
    def export(
        self,
        graph: DAGNodesGraph,
        filename: str,
        format: Optional[str] = None,
        **kwargs,
    ):
        pass

    def _check_format(self, filename: str, format: Optional[str] = None) -> str:
        if format is None:
            format = Path(filename).suffix[1:]
        if format not in self.exportable_formats():
            raise ValueError(f"Format {format} not supported")
        return format



import inspect
import typing as t
import pydantic as pyd
from pydantic.generics import GenericModel
from pipelime.utils.pydantic_types import TypeDef, CallableDef
from pipelime.stages import SampleStage
from pipelime.items import Item

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample

ActionTp = t.TypeVar("ActionTp", bound=t.Callable)


@t.overload
def register_action(
    *, title: t.Optional[str] = None, description: t.Optional[str] = None
) -> t.Callable[[ActionTp], ActionTp]:
    ...


@t.overload
def register_action(__action: ActionTp) -> ActionTp:
    ...


def register_action(
    __action=None, *, title: t.Optional[str] = None, description: t.Optional[str] = None
):
    """Register an action (class or function)

    Args:
        __action (_type_, optional): _description_. Defaults to None.
        title (t.Optional[str], optional): _description_. Defaults to None.
        description (t.Optional[str], optional): _description_. Defaults to None.
    """

    def _decorator(func):
        from pathlib import Path
        from pipelime.cli.utils import ActionInfo, PipelimeSymbolsHelper

        name = title or func.__name__
        try:
            source_path = inspect.getfile(func)
            class_path = Path(source_path).resolve().as_posix()

            if (func.__module__.replace(".", "/") + ".py") in class_path:
                class_path = func.__module__ + "." + func.__name__
            else:
                class_path = class_path + ":" + func.__name__
        except (TypeError, OSError):
            # this happens when the class does not come from a file
            class_path = "__main__." + func.__name__

        docs = description
        if docs is None:
            docs = inspect.getdoc(func) or ""

        PipelimeSymbolsHelper.register_action(
            name,
            ActionInfo(
                action=func,
                name=name,
                description=docs,
                classpath=class_path,
            ),
        )
        return func

    if __action is None:
        return _decorator
    return _decorator(__action)


ItTp = t.TypeVar("ItTp", bound=Item)
ValTp = t.TypeVar("ValTp")


class ParsedItem(
    GenericModel,
    t.Generic[ItTp, ValTp],
    extra="forbid",
    copy_on_model_validation="none",
    arbitrary_types_allowed=True,
):
    raw_item: ItTp
    parsed_value: ValTp

    def __call__(self) -> ValTp:
        return self.parsed_value

    @property
    def value(self) -> ValTp:
        return self.parsed_value

    @classmethod
    def raw_item_type(cls) -> t.Type[ItTp]:
        return cls.__fields__["raw_item"].outer_type_

    @classmethod
    def parsed_value_type(cls) -> t.Type[ValTp]:
        return cls.__fields__["parsed_value"].outer_type_

    @classmethod
    def value_to_item_data(cls, value) -> t.Any:
        if hasattr(value, "__to_item_data__"):
            return value.__to_item_data__()
        elif isinstance(value, pyd.BaseModel):
            return value.dict()
        return value

    @classmethod
    def make_raw_item(cls, value) -> ItTp:
        return cls.raw_item_type().make_new(cls.value_to_item_data(value))

    @classmethod
    def make_parsed_value(cls, value) -> ValTp:
        pvtp = cls.parsed_value_type()
        return (  # type: ignore
            pyd.parse_obj_as(pvtp, value)
            if issubclass(pvtp, pyd.BaseModel)
            else pvtp(value)
        )

    @classmethod
    def make_new(cls, value) -> ItTp:
        return cls.make_raw_item(value)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        elif isinstance(value, Item):
            if isinstance(value, cls.raw_item_type()):
                return cls(
                    raw_item=value,
                    parsed_value=cls.make_parsed_value(value()),
                )
        elif isinstance(value, cls.parsed_value_type()):
            return cls(
                raw_item=cls.make_raw_item(value),
                parsed_value=value,
            )
        else:
            try:
                value = cls.make_parsed_value(value)
            except Exception:
                pass
            else:
                return cls(
                    raw_item=cls.make_raw_item(value),
                    parsed_value=value,
                )
        raise TypeError(
            f"{value} is neither `{cls.raw_item_type()}` nor `{cls.parsed_value_type()}`"
        )


class ParsedData(ParsedItem[Item, ValTp], t.Generic[ValTp]):
    pass


class ModelDynamicKey:
    __slots__ = ("owner", "item_tp", "default", "default_factory", "extra_kwargs")

    def __init__(
        self,
        item_tp: t.Union[t.Type[Item], t.Type[ParsedItem]],
        default: t.Any = ...,
        default_factory: t.Optional[t.Callable[[], t.Any]] = None,
        extra_kwargs: t.Mapping[str, t.Any] = {},
    ):
        self.item_tp = item_tp
        self.default = default
        self.default_factory = default_factory
        self.extra_kwargs = extra_kwargs

    def validate(self, key) -> t.Union[t.Type[Item], t.Type[ParsedItem]]:
        class Config(pyd.BaseConfig):
            arbitrary_types_allowed = True

        parser_model = pyd.create_model(
            "DynamicKeyParser",
            __config__=Config,
            **{
                key: (
                    self.item_tp,
                    pyd.Field(
                        self.default,
                        default_factory=self.default_factory,
                        **self.extra_kwargs,
                    ),
                )
            },
        )
        parsed_values = parser_model.parse_obj(
            {key: getattr(self.owner, key)} if hasattr(self.owner, key) else {}
        )
        return getattr(parsed_values, key)


def DynamicKey(
    item_tp: t.Union[t.Type[Item], t.Type[ParsedItem]],
    default: t.Any = ...,
    *,
    default_factory: t.Optional[t.Callable[[], t.Any]] = None,
    **field_kwargs,
) -> t.Any:
    if default is not ... and default_factory is not None:
        raise ValueError("cannot specify both default and default_factory")

    return pyd.PrivateAttr(
        ModelDynamicKey(item_tp, default, default_factory, field_kwargs)
    )


DerivedEntityTp = t.TypeVar("DerivedEntityTp", bound="BaseEntity")


class BaseEntity(
    pyd.BaseModel,
    extra="allow",
    copy_on_model_validation="none",
    arbitrary_types_allowed=True,
):
    """The base class for all input/output entity models."""

    def __init__(self, **data):
        # create an item field from raw values
        # NB: if the field is optional, it can be None
        for k, v in data.items():
            if k in self.__fields__ and not isinstance(v, Item):
                k_field = self.__fields__[k]
                if issubclass(k_field.outer_type_, Item) and (
                    k_field.required or v is not None
                ):
                    data[k] = self.__fields__[k].outer_type_.make_new(v)
        super().__init__(**data)

        # assign self as the owner of dynamic key fields
        for k in self.__private_attributes__:
            v = getattr(self, k)
            if isinstance(v, ModelDynamicKey):
                v.owner = self

    def _iter(self, *args, **kwargs):
        # skip None fields and bypass ParsedItem
        for k, v in super()._iter(*args, **kwargs):
            if v is not None:
                if isinstance(v, t.Mapping) and "raw_item" in v:
                    if v["raw_item"] is not None:
                        yield k, v["raw_item"]
                else:
                    yield k, v

    @classmethod
    def merge(
        cls: t.Type[DerivedEntityTp], __other__: "BaseEntity", /, **kwargs
    ) -> DerivedEntityTp:
        """Creates a new entity by merging `other` entity with extra `kwargs` fields."""
        other_dict = __other__.dict()
        for k, v in kwargs.items():
            if not isinstance(v, Item):
                # NB: None is a valid value for optional fields
                my_k_field = cls.__fields__.get(k, None)
                if not my_k_field or my_k_field.required or v is not None:
                    # user has no preference on the actual Item class,
                    # so keep the original item type if it is a subclass
                    # of the declared output item type
                    if k in other_dict:
                        other_item_obj = other_dict[k]
                        if other_item_obj is not None:
                            my_item_cls = Item
                            if my_k_field:
                                my_item_cls = my_k_field.outer_type_
                                if issubclass(my_item_cls, ParsedItem):
                                    my_item_cls = my_item_cls.raw_item_type()
                            if isinstance(other_item_obj, my_item_cls):
                                kwargs[k] = other_item_obj.make_new(
                                    ParsedItem.value_to_item_data(v)
                                )
        return cls(**{**other_dict, **kwargs})


class ActionDef(CallableDef):
    """Action definition, can be a registered action title, a callable
    (cfr. Choixe `$call`/`$symbol`), a `class.path.to.a.function`,
    a `path/to/a/file.py:function` or a mapping where the key is like above,
    while the value is the list of `__init__` arguments
    (mapping, sequence or single value).
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_action

    @classmethod
    def validate_action(
        cls,
        value: t.Union[
            CallableDef, t.Callable, str, t.Mapping[t.Union[str, t.Callable], t.Any]
        ],
    ) -> "ActionDef":
        from pipelime.cli.utils import PipelimeSymbolsHelper

        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            # action is function
            act = PipelimeSymbolsHelper.get_action(value)
            if act is not None:
                value = act.action
        elif isinstance(value, t.Mapping):
            # action is a callable instance
            name, args = next(iter(value.items()))
            if isinstance(name, str):
                act = PipelimeSymbolsHelper.get_actions().get(name, None)
                if act is not None:
                    value = {act.action: args}
        return cls.validate(value)


class BaseEntityType(TypeDef[BaseEntity]):
    """An entity type. It accepts both type names and string."""


class EntityAction(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    """An action and its associated input entity model."""

    action: ActionDef = pyd.Field(
        ...,
        description=(
            "The action callable to run (can be a class path). The expected annotation "
            "is `(BaseEntitySubClass) -> BaseEntityLike`. Return type annotation "
            "is not mandatory."
        ),
    )
    input_type: BaseEntityType = pyd.Field(
        BaseEntity,
        description=(
            "The input type of the action (can be a string). If None, "
            "it is inferred from the callable's annotations or set to BaseEntity."
        ),
    )

    @pyd.validator("action")
    def validate_action(cls, v):
        params = v.full_signature.parameters
        if not params:
            raise ValueError(
                "The action must have at least one argument, ie, the input model entity."
            )
        first_param = next(iter(params.values()))
        if (
            first_param.kind == inspect.Parameter.KEYWORD_ONLY
            or first_param.kind == inspect.Parameter.VAR_KEYWORD
        ):
            raise ValueError(
                "The action must have at least one positional argument, "
                "ie, the input model entity."
            )
        return v

    @pyd.validator("input_type", always=True)
    def validate_input_type(cls, v, values):
        if "action" in values:  # if not True, an error should have been raised yet
            action_t = values["action"].args_type[0]
            if v is None:
                v = BaseEntityType.create(action_t or BaseEntity)
            else:
                if action_t is not None and not issubclass(v.value, action_t):
                    if not issubclass(action_t, v.value):
                        raise ValueError(
                            "Action annotation is incompatible with the input type."
                        )
                    v = BaseEntityType.create(action_t)  # use the more specific type

        return v

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        if isinstance(value, cls):
            return value
        if not isinstance(value, t.Mapping):
            value = {"action": value}
        if "action" not in value:
            value = {"action": value}
        return cls(**value)


class StageEntity(SampleStage, title="entity"):
    __root__: EntityAction = pyd.Field(..., description="The entity action to run.")

    def __init__(self, __root__: EntityAction, **data):
        super().__init__(__root__=__root__, **data)  # type: ignore

    def __call__(self, x: "Sample") -> "Sample":
        from pipelime.sequences import Sample

        return Sample(self.__root__.action(self.__root__.input_type.value(**x)).dict())



import subprocess
from string import Formatter
from typing import Any, Dict, Mapping, Sequence

from pydantic import Field

from pipelime.piper import PipelimeCommand, PiperPortType


class ShellCommand(PipelimeCommand, title="shell"):
    """A generic pipelime command wrapping a shell command."""

    command: str = Field(
        ...,
        alias="c",
        description=(
            "The shell command to execute. Use the `{name}` syntax to refer "
            "to specific inputs and outputs, then remaining inputs/outputs keys "
            "are appended as `--key value` arguments."
        ),
    )
    inputs: Dict[str, Any] = Field(
        default_factory=dict,
        alias="i",
        description=(
            "The input options. They will be matched with outputs "
            "of other commands when building a piper graph."
        ),
        piper_port=PiperPortType.INPUT,
    )
    outputs: Dict[str, Any] = Field(
        default_factory=dict,
        alias="o",
        description=(
            "The output options. They will be matched with inputs "
            "of other commands when building a piper graph."
        ),
        piper_port=PiperPortType.OUTPUT,
    )

    def get_inputs(self) -> Dict[str, Any]:
        return self.inputs

    def get_outputs(self) -> Dict[str, Any]:
        return self.outputs

    def command_name(self) -> str:
        return self.command

    def _to_command_chunk(self, key: str, value: Any) -> str:
        cmd = ""
        if isinstance(value, Sequence):
            cmd += f" --{key} {value[0]} {self._to_command_chunk(key, value[1:])}"
        elif isinstance(value, Mapping):
            raise NotImplementedError("Mapping values are not supported")
        elif isinstance(value, bool):
            if value:
                cmd += f" --{key}"
        else:
            cmd += f" --{key} {value}"
        return cmd

    def run(self) -> None:
        cmd = self.command
        fields = [fname for _, fname, _, _ in Formatter().parse(cmd) if fname]
        args = {**self.inputs, **self.outputs}
        cmd = cmd.format(**args)

        for key in set(args.keys()).difference(fields):
            value = args[key]
            cmd += self._to_command_chunk(key, value)

        subprocess.run(cmd, shell=True)



from typing import ContextManager


class ContextManagerList:
    def __init__(self, *args: ContextManager):
        self._cms = args

    def __enter__(self):
        for cm in self._cms:
            cm.__enter__()

    def __exit__(self, exc_type, exc_value, traceback):
        for cm in self._cms:
            cm.__exit__(exc_type, exc_value, traceback)


class CatchSignals:
    def __init__(self, signals=None):
        import signal
        self._signals = signals if signals is not None else [signal.SIGINT, signal.SIGTERM]
        self._original_handlers = {}
        self._interrupted = False
        self._released = False

    @property
    def interrupted(self):
        return self._interrupted

    def __enter__(self):
        import signal
        self._original_handlers = {}
        self._interrupted = False
        self._released = False

        for sig in self._signals:
            self._original_handlers[sig] = signal.getsignal(sig)
            signal.signal(sig, self.handler)

        return self

    def handler(self, signum, frame):
        self.release()
        self._interrupted = True

    def __exit__(self, type, value, tb):
        self.release()

    def release(self):
        import signal
        if self._released:
            return False

        for sig in self._signals:
            signal.signal(sig, self._original_handlers[sig])

        self._released = True
        return True



from pydantic import BaseModel


class OperationInfo(BaseModel, frozen=True):
    """Information on a running operation."""

    token: str
    """The piper token identifying the session"""

    node: str
    """The node currently being executed"""

    chunk: int
    """The current chunk number"""

    message: str = ""
    """A short description of the current chunk"""

    total: int
    """The total number of steps in the current chunk"""


class ProgressUpdate(BaseModel):
    """Advacncement of an operation."""

    op_info: OperationInfo
    """The operation info"""

    progress: int = 0
    """The progress of the current chunk"""

    finished: bool = False
    """Whether th operation has finished"""



from __future__ import annotations

from abc import ABC, abstractmethod
from math import prod
from typing import Any, Callable, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np


T = TypeVar("T", float, int, bool, Sequence[float], np.ndarray)


class RealFn(ABC):
    def __call__(self, x: T) -> T:
        if isinstance(x, (float, int, bool)):
            return self._call(np.array([float(x)])).tolist()[0]
        elif isinstance(x, Sequence) and not isinstance(x, str):
            return self._call(np.array(x)).tolist()
        elif isinstance(x, np.ndarray):
            return self._call(x)
        raise ValueError(f"Unsupported type {type(x)}")

    def __repr__(self) -> str:
        return f"f(x) = {self._repr()}"

    @abstractmethod
    def _call(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def integrate(self) -> RealFn:
        pass

    @abstractmethod
    def invert(self, start: float, stop: float) -> Optional[RealFn]:
        pass

    @abstractmethod
    def shift(self, dx: float, dy: float) -> RealFn:
        pass

    @abstractmethod
    def _repr(self) -> str:
        pass


t_fn_callable = Callable[[np.ndarray], np.ndarray]
t_fn = Union[float, Tuple[float, float], t_fn_callable, RealFn]
t_fn_pw_simple = Sequence[float]
t_fn_pw_ranges = Sequence[Tuple[float, t_fn]]
t_pw = Union[t_fn_pw_simple, t_fn_pw_ranges]
t_pdf = Union[RealFn, t_fn_callable, t_fn_pw_simple, t_fn_pw_ranges, str]


class LinearFn(RealFn):
    def __init__(self, m: float, q: float) -> None:
        super().__init__()
        self.m = m
        self.q = q

    def _call(self, x: np.ndarray) -> np.ndarray:
        return self.m * x + self.q

    def integrate(self) -> RealFn:
        if self.m == 0:
            return LinearFn(self.q, 0)
        return QuadraticFn(self.m / 2, self.q, 0)

    def invert(self, start: float, stop: float) -> Optional[RealFn]:
        if self.m == 0:
            return None
        return LinearFn(1 / self.m, -self.q / self.m)

    def shift(self, dx: float, dy: float) -> LinearFn:
        return LinearFn(self.m, self.q + dy - self.m * dx)

    def _repr(self) -> str:
        return f"{round(self.m, 2)}x + {round(self.q, 2)}"


class ConstantFn(LinearFn):
    def __init__(self, c: float) -> None:
        super().__init__(0, c)

    def _repr(self) -> str:
        return str(round(self.q, 2))

    def integrate(self) -> RealFn:
        return LinearFn(self.q, 0)

    def shift(self, dx: float, dy: float) -> ConstantFn:
        return ConstantFn(self.q + dy)


class QuadraticFn(RealFn):
    def __init__(self, a: float, b: float, c: float) -> None:
        super().__init__()
        self.a = a
        self.b = b
        self.c = c

    def _call(self, x: np.ndarray) -> np.ndarray:
        return self.a * x**2 + self.b * x + self.c

    def integrate(self) -> RealFn:
        return GenericFn(
            lambda x: self.a * x**3 / 3 + self.b * x**2 / 2 + self.c * x,
            -np.inf,
            np.inf,
        )

    def invert(self, start: float, stop: float) -> Optional[RealFn]:
        a, b, c, eps = self.a, self.b, self.c, 1e-6
        if a == 0:
            return LinearFn(b, c).invert(start, stop)
        vertex_x = -b / (2 * a)
        if start + eps < vertex_x < stop - eps:
            return None
        sv = -1 if (start + stop) / 2 < vertex_x else 1
        sa = -1 if a < 0 else 1
        p, q = self(np.array([start, stop])).tolist()
        return GenericFn(
            lambda x: (
                -b + sv * sa * np.sqrt(np.clip(b**2 - 4 * a * (c - x), 0, None))
            )
            / (2 * a),
            min(p, q),
            max(p, q),
        )

    def shift(self, dx: float, dy: float) -> QuadraticFn:
        a = self.a
        b = self.b - 2 * self.a * dx
        c = self.a * dx**2 - self.b * dx + self.c + dy
        return QuadraticFn(a, b, c)

    def _repr(self) -> str:
        return f"{round(self.a, 2)}x² + {round(self.b, 2)}x + {round(self.c, 2)}"


class GenericFn(RealFn):
    def __init__(
        self,
        fn: Callable[[np.ndarray], np.ndarray],
        start: float,
        stop: float,
        steps: int = 100,
    ) -> None:
        super().__init__()
        self.fn = fn
        self.start = start
        self.stop = stop
        self.steps = steps

    def _call(self, x: np.ndarray) -> np.ndarray:
        return self.fn(x)

    def _to_piecewise(self, start=None, stop=None) -> PiecewiseFn:
        start, stop = start or self.start, stop or self.stop
        x = np.linspace(start, stop, self.steps + 1, endpoint=True)
        y = self(x)
        x, y = x.tolist(), y.tolist()
        segments = [(x[i], (y[i], y[i + 1])) for i in range(len(x) - 1)]
        return PiecewiseFn.parse(segments, start, stop)

    def integrate(self) -> PiecewiseFn:
        return self._to_piecewise().integrate()

    def invert(self, start: float, stop: float) -> Optional[PiecewiseFn]:
        return self._to_piecewise(start=start, stop=stop).invert(start, stop)

    def shift(self, dx: float, dy: float) -> GenericFn:
        return GenericFn(lambda x: self(x - dx) + dy, self.start - dx, self.stop - dx)

    def _repr(self) -> str:
        return "code"


class PiecewiseFn(RealFn):
    @classmethod
    def parse_piece(cls, piece: t_fn, start: float, stop: float) -> RealFn:
        if isinstance(piece, RealFn):
            return piece
        elif isinstance(piece, float):
            return ConstantFn(piece)
        elif isinstance(piece, (tuple, list)):
            m = (piece[1] - piece[0]) / (stop - start)
            q = piece[0] - m * start
            return LinearFn(m, q)
        elif callable(piece):
            return GenericFn(piece, start, stop)
        else:
            raise TypeError("Piece must be a parsable function")

    @classmethod
    def parse(
        cls,
        fn: t_pw,
        start: Optional[float] = None,
        stop: Optional[float] = None,
    ) -> PiecewiseFn:
        if len(fn) == 0:
            raise ValueError("Piecewise function must have at least one piece")
        if all(isinstance(piece, float) for piece in fn):
            segments = {}
            start_, stop_ = start or 0.0, stop or 1.0
            cur = start_
            step = (stop_ - start_) / len(fn)
            for piece in fn:
                segments[cur] = cls.parse_piece(piece, cur, cur + step)  # type: ignore
                cur += step
            segments[stop_] = ConstantFn(0.0)
        elif all(isinstance(piece, (tuple, list)) for piece in fn):
            segments = {}
            keysteps = [x[0] for x in fn]  # type: ignore
            start_, stop_ = start or keysteps[0], stop or keysteps[-1]
            for i in range(len(keysteps)):
                cur, nxt = (
                    keysteps[i],
                    keysteps[i + 1] if i < len(keysteps) - 1 else stop_,
                )
                if cur == nxt:
                    continue
                segments[cur] = cls.parse_piece(fn[i][1], cur, nxt)  # type: ignore
            segments[stop_] = ConstantFn(0.0)
        else:
            raise TypeError("Function must be a list of all floats or all tuples")
        return cls(segments)

    def __init__(self, segments: Mapping[float, RealFn]) -> None:
        super().__init__()
        assert len(segments) > 0, "Piecewise function must have at least one piece"
        self.segments = segments
        self.keysteps = sorted(list(segments.keys()))

    def _call(self, x: np.ndarray) -> np.ndarray:
        results = np.zeros_like(x)
        for i in range(len(self.keysteps)):
            cur, nxt = (
                self.keysteps[i],
                self.keysteps[i + 1] if i < len(self.keysteps) - 1 else float("inf"),
            )
            mask = (cur <= x) & (x <= nxt)
            results[mask] = self.segments[cur](x[mask])
        return results

    def integrate(self) -> PiecewiseFn:
        segments = {}
        cumsum = 0.0
        for i in range(len(self.keysteps)):
            cur, nxt = (
                self.keysteps[i],
                self.keysteps[i + 1] if i < len(self.keysteps) - 1 else 0,
            )
            integral = self.segments[cur].integrate()
            c0 = cumsum - integral(cur)
            segments[cur] = integral.shift(0, c0)
            cumsum += integral(nxt) - integral(cur)

        segments[float("inf")] = ConstantFn(cumsum)
        integral = PiecewiseFn(segments)
        return integral.shift(0, -integral(0))

    def invert(self, start: float, stop: float) -> Optional[PiecewiseFn]:
        segments = {}
        last_start = last_stop = -float("inf")
        for i in range(len(self.keysteps)):  # pragma: no branch
            cur, nxt = (
                self.keysteps[i],
                self.keysteps[i + 1] if i < len(self.keysteps) - 1 else stop,
            )
            if cur <= start < nxt:
                fn = self.segments[cur]
                inverted = fn.invert(start, min(nxt, stop))
                is_const = isinstance(fn, LinearFn) and fn.m == 0

                # Piece not invertible -> no solution
                if inverted is None and not is_const:
                    return None

                # Compute the domain of the inverted function
                inverse_start, inverse_stop = fn(start), fn(min(nxt, stop))

                # If lower than previous segment or decreasing, no solution
                if (
                    inverse_start + 1e-5 < last_stop
                    or inverse_stop + 1e-5 < inverse_start
                ):
                    return None

                last_start = inverse_start
                last_stop = inverse_stop
                if inverted is not None:
                    segments[last_start] = inverted
                start = nxt

                # Break if we've reached the end of the domain
                if nxt >= stop:
                    break

        segments[float("inf")] = ConstantFn(last_start)
        return PiecewiseFn(segments)

    def shift(self, dx: float, dy: float) -> PiecewiseFn:
        return PiecewiseFn({k + dx: v.shift(dx, dy) for k, v in self.segments.items()})

    def _repr(self) -> str:
        return "piecewise"


class Distribution:
    def __init__(self, fn: RealFn, start: float, stop: float) -> None:
        self.start = start
        self.stop = stop

        primitive = fn.integrate()
        primitive = primitive.shift(0, -primitive(start))
        inverse_primitive = primitive.invert(start, stop)
        sum_ = primitive(stop) - primitive(start)

        assert inverse_primitive is not None, "The function must be invertible"

        self.pdf = lambda x: fn(np.clip(x, start, stop)) / sum_
        self.cdf = lambda x: primitive(x) / sum_
        self.inverse_cdf = lambda x: inverse_primitive(x * sum_)  # type: ignore

    def sample(self, n: int) -> np.ndarray:
        return self.inverse_cdf(np.random.rand(n))

    @classmethod
    def parse(
        cls, fn: t_pdf, start: Optional[float] = None, stop: Optional[float] = None
    ) -> Distribution:
        if isinstance(fn, str):
            raise NotImplementedError("Parsing strings is not implemented yet")
        if isinstance(fn, RealFn):
            return Distribution(fn, start or 0.0, stop or 1.0)
        elif callable(fn):
            start_, stop_ = start or 0.0, stop or 1.0
            rfn = PiecewiseFn.parse_piece(fn, start_, stop_)  # type: ignore
            return Distribution(rfn, start_, stop_)
        elif isinstance(fn, Sequence):
            rfn = PiecewiseFn.parse(fn, start=start, stop=stop)
            keysteps = sorted(list(rfn.segments.keys()))
            start_, stop_ = keysteps[0], keysteps[-1]
            return Distribution(rfn, start_, stop_)
        else:
            raise ValueError("Invalid distribution")


def _rand(
    start: Optional[float] = None,
    stop: Optional[float] = None,
    n: Union[int, Sequence[int]] = 0,
    pdf: Optional[t_pdf] = None,
    integer: bool = False,
) -> Union[float, Sequence[float]]:
    if pdf is None:
        start_, stop_, n_ = start or 0.0, stop or 1.0, n or 1
        results = np.random.uniform(start_, stop_, n_)
    else:
        distribution = Distribution.parse(pdf, start=start, stop=stop)
        n_ = prod(n) if isinstance(n, Sequence) else (n or 1)
        results = distribution.sample(n_)

    if integer:
        results = np.floor(results).astype(np.int64)

    if n == 0:
        return results.item()
    elif isinstance(n, int):
        return results.tolist()
    else:
        return results.reshape(n)  # type: ignore


def rand(*args, n: Union[int, Sequence[int]] = 0, pdf: Optional[t_pdf] = None) -> Any:
    if len(args) == 0:
        return _rand(n=n, pdf=pdf)
    elif len(args) == 1:
        a = args[0]
        return _rand(stop=a, n=n, pdf=pdf, integer=isinstance(a, int))
    elif len(args) == 2:
        a, b = args
        integer = isinstance(a, int) and isinstance(b, int)
        return _rand(start=a, stop=b, n=n, pdf=pdf, integer=integer)

    raise ValueError("Invalid arguments")



import json
import yaml
import toml
from io import TextIOWrapper

import typing as t

from pipelime.items import Item
from pipelime.items.base import deferred_classattr

_metadata_type = t.Union[str, int, float, bool, None, t.Sequence, t.Mapping]


class MetadataItem(Item[_metadata_type]):
    """A common base for items dealing with metadata."""

    @deferred_classattr
    def default_concrete(cls):
        return YamlMetadataItem

    @classmethod
    def validate(cls, raw_data: t.Any) -> _metadata_type:
        if not isinstance(
            raw_data, (str, int, float, bool, type(None), t.Sequence, t.Mapping)
        ):
            raise ValueError(  # pragma: no cover
                f"{cls}: raw data must be one of"
                " (str, int, float, bool, None, Sequence, Mapping)."
            )
        return raw_data

    @classmethod
    def pl_pretty_data(cls, value: _metadata_type) -> t.Any:
        from rich.pretty import Pretty

        return Pretty(value, indent_guides=True, expand_all=True)


class JsonMetadataItem(MetadataItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".json",)

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> _metadata_type:
        return json.load(TextIOWrapper(fp))

    @classmethod
    def encode(cls, value: _metadata_type, fp: t.BinaryIO):
        json.dump(value, TextIOWrapper(fp))


class YamlMetadataItem(MetadataItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".yaml", ".yml")

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> _metadata_type:
        return yaml.safe_load(TextIOWrapper(fp))

    @classmethod
    def encode(cls, value: _metadata_type, fp: t.BinaryIO):
        yaml.safe_dump(value, TextIOWrapper(fp), sort_keys=False)


class TomlMetadataItem(MetadataItem):
    @classmethod
    def file_extensions(cls) -> t.Sequence[str]:
        return (".toml", ".tml")

    @classmethod
    def decode(cls, fp: t.BinaryIO) -> t.Mapping:
        return toml.load(TextIOWrapper(fp))

    @classmethod
    def encode(cls, value: t.Mapping, fp: t.BinaryIO):
        toml.dump(value, TextIOWrapper(fp))

    @classmethod
    def validate(cls, raw_data: t.Any) -> _metadata_type:
        if not isinstance(raw_data, t.Mapping):
            raise ValueError(f"{cls}: raw data must be a Mapping")  # pragma: no cover
        return raw_data



from __future__ import annotations
from abc import ABC, abstractmethod

import pydantic as pyd
import typing as t
import inspect

if t.TYPE_CHECKING:
    from pipelime.sequences import Sample


class SampleStage(pyd.BaseModel, ABC, extra="forbid", copy_on_model_validation="none"):
    """Base class for all sample stages."""

    @abstractmethod
    def __call__(self, x: "Sample") -> "Sample":
        pass

    def __rshift__(self, other: SampleStage) -> SampleStage:
        """`>>` composes two stages: `other` after `self`."""
        return StageCompose([self, other])

    def __lshift__(self, other: SampleStage) -> SampleStage:
        """`<<` composes two stages: `other` before `self`."""
        return StageCompose([other, self])


class StageIdentity(SampleStage, title="identity"):
    """Returns the input sample."""

    def __call__(self, x: "Sample") -> "Sample":
        return x


class StageLambda(SampleStage, title="lambda"):
    """Applies a callable to the sample."""

    func: t.Callable = pyd.Field(
        ...,
        description="The callable to apply, accepting a Sample and returning a Sample.",
    )

    def __init__(self, func, **data):
        super().__init__(func=func, **data)  # type: ignore

    def __call__(self, x: "Sample") -> "Sample":
        return self.func(x)


class StageInput(pyd.BaseModel, extra="forbid", copy_on_model_validation="none"):
    """A stage is SampleStage object, `<name>` or `<name>: <args>` mapping,
    where `<name>` is `compose`, `remap`, `albumentations` etc,
    while `<args>` is a mapping of its arguments."""

    __root__: SampleStage

    def __call__(self, x: "Sample") -> "Sample":
        return self.__root__(x)

    def __str__(self) -> str:
        return str(self.__root__)

    def __repr__(self) -> str:
        return repr(self.__root__)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        from pipelime.cli.utils import create_stage_from_config

        if isinstance(value, StageInput):
            return value
        if isinstance(value, SampleStage):
            return StageInput(__root__=value)
        if isinstance(value, (str, bytes)):
            return StageInput(__root__=create_stage_from_config(str(value), None))
        if isinstance(value, t.Mapping):
            return StageInput(
                __root__=create_stage_from_config(*next(iter(value.items())))
            )
        raise ValueError(f"Invalid stage definition: {value}")


class StageCompose(SampleStage, title="compose"):
    """Applies a sequence of stages."""

    stages: t.Sequence[StageInput] = pyd.Field(
        ...,
        description="The stages to apply. " + str(inspect.getdoc(StageInput)),
    )

    def __init__(
        self,
        stages: t.Sequence[
            t.Union[SampleStage, str, t.Mapping[str, t.Mapping[str, t.Any]]]
        ],
        **data,
    ):
        super().__init__(stages=stages, **data)  # type: ignore

    def __call__(self, x: "Sample") -> "Sample":
        for s in self.stages:
            x = s(x)  # type: ignore
        return x



import importlib
import importlib.util
import os
import sys
from contextlib import ContextDecorator
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, Union
from uuid import uuid1


class add_to_sys_path(ContextDecorator):
    """add_to_sys_path context decorator temporarily adds a path to sys.path"""

    def __init__(self, path: str) -> None:
        self._new_cwd = path

    def __enter__(self) -> None:
        sys.path.insert(0, self._new_cwd)

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        sys.path.pop(0)


class add_to_sys_module(ContextDecorator):
    """
    ☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️
    THIS IS A HACK TO ALLOW DYNAMIC IMPORT FROM FILES CONTAINING DATACLASSES.
    THE USAGE OF THIS DECORATOR SHOULD BE AVOIDED AS IT COULD POTENTIALLY BREAK STUFF.

    THIS DECORATOR IS CURRENTLY USED ONLY IN THIS MODULE AND WE PLAN TO REMOVE IT AS
    SOON AS A BETTER SOLUTION IS FOUND.
    ☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️☢️

    add_to_sys_module context decorator temporarily adds a module to sys.modules
    """

    def __init__(self, module: ModuleType, id_: str) -> None:
        self._module = module
        self._id_ = id_

    def __enter__(self) -> None:
        if self._id_ in sys.modules:  # pragma: no cover
            raise RuntimeError(f"Attempted to overwrite system module {self._id_}")

        sys.modules[self._id_] = self._module

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        sys.modules.pop(self._id_)


def import_module_from_file(
    module_file_path: Union[str, Path], cwd: Optional[Path] = None
) -> ModuleType:
    """Import a python module from a file.

    Args:
        module_file_path (Union[str, Path]): the path to the `.py` module file.
        cwd (Optional[Path], optional): the folder to use for relative module import.
            If None, the file parent folder will be used. Defaults to None.

    Raises:
        ImportError: _description_

    Returns:
        ModuleType: _description_
    """
    module_path = Path(module_file_path)
    if not module_path.is_absolute() and cwd is not None:
        module_path = cwd / module_path

    with add_to_sys_path(str(module_path.parent)):
        return import_module_from_class_path(module_path.stem)


def import_module_from_class_path(module_class_path: str) -> ModuleType:
    return importlib.import_module(module_class_path)


def import_module(
    module_file_or_class_path: str, cwd: Optional[Path] = None
) -> ModuleType:
    return (
        import_module_from_file(module_file_or_class_path, cwd)
        if module_file_or_class_path.endswith(".py")
        else import_module_from_class_path(module_file_or_class_path)
    )


def import_symbol(symbol_path: str, cwd: Optional[Path] = None) -> Any:
    """Dynamically imports a given symbol. A symbol can be either:

    - A filesystem path followed by the name of an object to import, like:
        "path/to/my_file.py:MyClass"

    - A python module path, like "module.submodule.MyClass"

    Anything can be imported: modules, classes, functions, etc...

    Args:
        symbol_path (str): The symbol to import
        cwd (Optional[Path], optional): The current working directory to resolve
            relative imports when loading from file. If None, the file parent folder
            will be used. Defaults to None.

    Raises:
        ImportError: If anything goes wrong with the import.

    Returns:
        Any: The dynamically imported object.
    """
    try:
        if ":" in symbol_path:
            # path/to/my_file.py:MyClass
            module_path, _, symbol_name = symbol_path.rpartition(":")
            module_ = import_module_from_file(module_path, cwd)
            symbol_name = symbol_name.split(".")
        else:
            # package.module.MyClass.MyNestedClass
            module_path, _, symbol_name = symbol_path.rpartition(".")
            symbol_name = [symbol_name]
            module_ = None
            while module_path:
                try:
                    module_ = import_module_from_class_path(module_path)
                    module_path = None
                except ModuleNotFoundError:
                    # the symbol is nested, so we need to import the parent class path
                    module_path, _, cl_path = module_path.rpartition(".")
                    symbol_name.insert(0, cl_path)

        if module_ is None:
            raise ModuleNotFoundError("Module path not found")

        # import the symbol
        # if nested, we need to import the parent symbols first
        symbol = module_
        for name in symbol_name:
            symbol = getattr(symbol, name)
        return symbol
    except Exception as e:
        raise ImportError(f"Cannot import {symbol_path}") from e



from typing import Optional

class MyCaller:
    def __init__(self, steps_back: int = 0):
        """Get information about your caller.

        Args:
            steps_back (int, optional): use 1 to get the caller of your caller, 2 to get
                the grandparent of your caller, etc. Defaults to 0.
        """
        import inspect

        self._caller = inspect.stack()[2+steps_back]

    @property
    def globals(self):
        return self._caller.frame.f_globals

    @property
    def locals(self):
        return self._caller.frame.f_locals

    @property
    def filename(self) -> str:
        return self._caller.filename

    @property
    def module(self) -> str:
        return self._caller.frame.f_globals["__name__"]

    @property
    def package(self) -> str:
        return self.module.partition(".")[0]

    @property
    def docstrings(self) -> Optional[str]:
        import inspect

        return inspect.getdoc(self._caller.frame.f_globals.get(self._caller.function))
